/****************************************************************************
 * standard net-snmp include files
 ****************************************************************************/
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

/****************************************************************************
 * Other mib implementation include files
 ****************************************************************************/
//#include "../status_modem_ant.h"

/****************************************************************************
 * adsl module include files
 ****************************************************************************/
#include "adsl.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <asm/tc3162/tc3162dmtctrl.h>

//#include "adsl_enums.h"

static void
initialize_table_adslChannelTable(void);

static void
initialize_table_adslChannelTable2(void);

/****************************************************************************
 * external routines
 ****************************************************************************/
//extern int Interface_Index_By_Name ( char *, int);

/****************************************************************************
 * private routines
 ****************************************************************************/
FILE* TTYS0;
#define DPRINTF(args...) \
	TTYS0=fopen("/dev/ttyS0","w"); \
	fprintf(TTYS0,args); \
	fclose(TTYS0) 


static void getDownUpStreamRate(int *downlinkspeed,int *uplinkspeed)
{
    DMTCTRL_CONNECTION_INFO Arg;
    DMTCTRL_DMT_INFO ArgStatus;

    int nFd = open( "/dev/dmtdev", O_RDWR );
    if( nFd == -1  )
      {
	DPRINTF("getDownUpStreamRate:severe error for open /dev/dmtdev for ADSL device\n");
	return;
      }


    ioctl( nFd, DMTIOCTL_GET_DMT_INFO, &ArgStatus );
    if(ArgStatus.dmt_status!=ADSL_UP)
      {
	close(nFd);
	*downlinkspeed = 0;
	*uplinkspeed = 0;
	return;
      }

    
    ioctl( nFd, DMTIOCTL_GET_CONNECTION_INFO, &Arg );
        
    //DPRINTF("near-end interleaved channel bit rate: %d kbps\n",Arg.actualBitrateNearInt);
    //DPRINTF("near-end fast channel bit rate: %d kbps\n",Arg.actualBitrateNearFast);
    //DPRINTF("far-end interleaved channel bit rate: %d kbps\n",Arg.actualBitrateFarInt);
    //DPRINTF("far-end fast channel bit rate: %d kbps\n",Arg.actualBitrateFarFast);
    if(Arg.actualBitrateNearInt)
      *downlinkspeed=Arg.actualBitrateNearInt;
    if(Arg.actualBitrateNearFast)
      *downlinkspeed=Arg.actualBitrateNearFast;
    if(Arg.actualBitrateFarInt)
      *uplinkspeed=Arg.actualBitrateFarInt;
    if(Arg.actualBitrateFarFast)
      *uplinkspeed=Arg.actualBitrateFarFast;
    close(nFd);      
}

/****************************************************************************
 * private variables 
 ****************************************************************************/
//static long     adslLinkStatus = 0;     /* XXX: set default value */
//static long     adslPhyIfIndex = 0;     /* XXX: set default value */

/****************************************************************************
 * public (exported) variables
 ****************************************************************************/
netsnmp_table_data_set *adslChannelTable_table_set;
netsnmp_table_data_set *adslChannelTable_table_set2;

/****************************************************************************
 * Function Name  : init_adsl
 * Arguments      :
 * Returns        : 
 * Variables      : 
 * Modifies       :  
 * Description    : Initializes the adsl MIB module
 *                  here we initialize all the scalar
 ****************************************************************************/
void
init_adsl(void)
{
    /*
     * a debugging statement.  Run the agent with -Dadsl to see
     * the output of this debugging statement. 
     */
    DEBUGMSGTL(("adsl", "Initializing the adsl module\n"));
    
    /*
     * adslChannelTable Table initialization 
     */
    initialize_table_adslChannelTable();
    initialize_table_adslChannelTable2();
    DEBUGMSGTL(("adsl", "Done initalizing adsl module\n"));
}

/****************************************************************************
 * Table Initializators
 ***************************************************************************/


/****************************************************************************
 * Function Name  : initialize_table_adslChannelTable
 * Arguments      :
 * Returns        : 
 * Variables      : 
 * Modifies       :  
 * Description    : Initialize the adslChannelTable table by defining its 
 *                  contents and how it's structured
 ****************************************************************************/
static void
initialize_table_adslChannelTable(void)
{
  //    static oid      adslChannelTable_oid[] =
  //      { 1, 3, 6, 1, 4, 1, 1372, 100, 1, 2, 1, 3 };
  static oid      adslChannelTable_oid[] =
          { 1, 3, 6, 1, 2, 1,   10,  94, 1, 1, 4};

  size_t          adslChannelTable_oid_len =
    OID_LENGTH(adslChannelTable_oid);
  netsnmp_table_data_set *table_set;


  /*
   * create the table structure itself 
   */
  table_set = netsnmp_create_table_data_set("adslChannelTable");
  table_set->allow_creation = 0;              /* no creation ALLOWED */
  table_set->table->store_indexes = 1;        /* needed for table handling */

  /*
   * Adding indexes to the table
   */
#if 0
  DEBUGMSGTL(("adsl",
	      "adding indexes to table adslChannelTable\n"));
  netsnmp_table_helper_add_indexes((netsnmp_table_registration_info *)table_set->table, 
				   ASN_INTEGER, 
				   0);
#endif

  /* index: adslChannelIfIndex */

  DEBUGMSGTL(("adsl",
	      "adding column types to table adslChannelTable\n"));
#if 0
  netsnmp_table_set_multi_add_default_row(table_set,
  					  COLUMN_ADSLCHANNELIFINDEX,
  					  ASN_INTEGER, 0, NULL, 0,
					  COLUMN_ADSLLINETYPE,
					  ASN_INTEGER, 0, NULL, 0,
					  COLUMN_ADSLUPSTREAMRATE,
					  ASN_GAUGE, 0, NULL, 0,
					  COLUMN_ADSLDOWNSTREAMRATE,
					  ASN_GAUGE, 0, NULL, 0,
					  COLUMN_ADSLUPSTREAMSNR,
					  ASN_OPAQUE_FLOAT, 0, NULL, 0,
					  COLUMN_ADSLDOWNSTREAMSNR,
					  ASN_OPAQUE_FLOAT, 0, NULL, 0,
					  0);
#endif

  netsnmp_table_dataset_add_index(table_set, ASN_INTEGER);
  netsnmp_table_set_multi_add_default_row(table_set,
					  1,
  					  ASN_INTEGER, 0, NULL, 0,
					  2,
					  ASN_INTEGER, 0, NULL, 0,
					  3,
					  ASN_INTEGER, 0, NULL, 0,
					  4,
					  ASN_INTEGER, 0, NULL, 0,
					  0);

  /*
   * registering the table 
   * adslChannelTable_handler not needed since update is done by pooling timer
   * in lsilAdslMib.c 
   */
  netsnmp_register_table_data_set(netsnmp_create_handler_registration
				  ("adslChannelTable",
				   NULL, 
				   adslChannelTable_oid,
				   adslChannelTable_oid_len,
				   HANDLER_CAN_RWRITE), table_set, NULL);

  /*
   * store the table for general use 
   */
  adslChannelTable_table_set = table_set;

  adslGroup_sync(0,NULL);

  snmp_alarm_register(5, 
		      SA_REPEAT,  
		      adslGroup_sync,
		      NULL
		      );
}

/***************************************************************************
 * SYSTEM TABLE SYNC
 ***************************************************************************/

/*************************************************************************
 * CUSTOM TABLE HANDLERS
 *************************************************************************/
 
/****************************************************************************
 * Function Name  : adslGroup_sync
 * Arguments      : 
 * Returns        : 
 * Variables      : 
 * Modifies       :  
 * Description    : sync dnmt status and ADSL group
 ****************************************************************************/
static int s_downrate=0;
static int found_table=0;
void 
adslGroup_sync (unsigned int clientreg, void *clientarg)
{
  //  static time_t lastModemAntOamChange;
  int uprate, downrate;

  //    if(fileChanged(MODEM_ANT_STATUS_FILE, &lastModemAntOamChange))
  {
    //  adslChannelTable_sync();
  }
  getDownUpStreamRate(&downrate,&uprate);
  //DPRINTF("down %d up %d\n",downrate, uprate);
  s_downrate=downrate*1000; // bps
  adslChannelTable_sync();
}

void adslChannelTable_sync()
{
  int  ifIndex, auxValue;// foundRow=0;
  //float auxFloat=0;
  netsnmp_table_data_set *table_set = adslChannelTable_table_set;
  netsnmp_table_row *row;
  //  netsnmp_table_data_set_storage *auxData;

  //  DEBUGMSGTL(("adsl", "adslChannelTable_sync create new row adslInterleave0 %d\n", ifIndex));
  //row = netsnmp_table_data_set_create_row_from_defaults(table_set->default_row);
  if( found_table == 0) {
    row = netsnmp_create_table_data_row();
    ifIndex = 8;
    netsnmp_table_row_add_index(row, ASN_INTEGER, (char *)&ifIndex, 4);
  } else {
    row = table_set->table->first_row;
  }
  //netsnmp_table_dataset_add_row(table_set, row);  /* add row to the table */
  ifIndex = 11;
  netsnmp_set_row_column(row, 1, ASN_INTEGER, (char *)(&ifIndex), 4); // 1
  //auxValue = ADSLLINETYPE_INTERLEAVEDONLY;
  //auxValue = 12;
  //auxValue = downrate;
  //auxValue = 65536;
  //auxValue = s_downrate;
  //netsnmp_set_row_column(row, 2, ASN_INTEGER, (char *)(&auxValue), 4); // 2
  netsnmp_set_row_column(row, 2, ASN_INTEGER, (char *)(&s_downrate), 4); // 2
  //auxValue = modem_ant_get_InlUpStreamSpeed(0);
  auxValue = 13;
  netsnmp_set_row_column(row, 3, ASN_INTEGER, (char *)(&auxValue), 4); // 3
  //auxValue = modem_ant_get_InlDnStreamSpeed(0);
  auxValue = 14;
  netsnmp_set_row_column(row, 4, ASN_INTEGER, (char *)(&auxValue), 4); // 4

  if( found_table == 0) {
    netsnmp_table_dataset_add_row(table_set, row);  /* add row to the table */
    found_table = 1;
  }
}

#if 1
/* ===================================================================================== */
static void
initialize_table_adslChannelTable2(void)
{
  //    static oid      adslChannelTable_oid[] =
  //      { 1, 3, 6, 1, 4, 1, 1372, 100, 1, 2, 1, 3 };
  static oid      adslChannelTable_oid[] =
          { 1, 3, 6, 1, 2, 1,   10,  94, 1, 1, 5};

  size_t          adslChannelTable_oid_len =
    OID_LENGTH(adslChannelTable_oid);
  netsnmp_table_data_set *table_set;


  /*
   * create the table structure itself 
   */
  table_set = netsnmp_create_table_data_set("adslChannelTable2");
  table_set->allow_creation = 0;              /* no creation ALLOWED */
  table_set->table->store_indexes = 1;        /* needed for table handling */

  /*
   * Adding indexes to the table
   */
#if 0
  DEBUGMSGTL(("adsl",
	      "adding indexes to table adslChannelTable\n"));
  netsnmp_table_helper_add_indexes((netsnmp_table_registration_info *)table_set->table, 
				   ASN_INTEGER, 
				   0);
#endif

  /* index: adslChannelIfIndex */

  DEBUGMSGTL(("adsl",
	      "adding column types to table adslChannelTable\n"));
#if 0
  netsnmp_table_set_multi_add_default_row(table_set,
  					  COLUMN_ADSLCHANNELIFINDEX,
  					  ASN_INTEGER, 0, NULL, 0,
					  COLUMN_ADSLLINETYPE,
					  ASN_INTEGER, 0, NULL, 0,
					  COLUMN_ADSLUPSTREAMRATE,
					  ASN_GAUGE, 0, NULL, 0,
					  COLUMN_ADSLDOWNSTREAMRATE,
					  ASN_GAUGE, 0, NULL, 0,
					  COLUMN_ADSLUPSTREAMSNR,
					  ASN_OPAQUE_FLOAT, 0, NULL, 0,
					  COLUMN_ADSLDOWNSTREAMSNR,
					  ASN_OPAQUE_FLOAT, 0, NULL, 0,
					  0);
#endif

  netsnmp_table_dataset_add_index(table_set, ASN_INTEGER);
  netsnmp_table_set_multi_add_default_row(table_set,
					  1,
  					  ASN_INTEGER, 0, NULL, 0,
					  2,
					  ASN_INTEGER, 0, NULL, 0,
					  3,
					  ASN_INTEGER, 0, NULL, 0,
					  4,
					  ASN_INTEGER, 0, NULL, 0,
					  0);

  /*
   * registering the table 
   * adslChannelTable_handler not needed since update is done by pooling timer
   * in lsilAdslMib.c 
   */
  netsnmp_register_table_data_set(netsnmp_create_handler_registration
				  ("adslChannelTable2",
				   NULL, 
				   adslChannelTable_oid,
				   adslChannelTable_oid_len,
				   HANDLER_CAN_RWRITE), table_set, NULL);

  /*
   * store the table for general use 
   */
  adslChannelTable_table_set2 = table_set;

  adslGroup_sync2(0,NULL);

  snmp_alarm_register(5, 
  		      SA_REPEAT,  
  		      adslGroup_sync2,
  		      NULL
  		      );
}

/***************************************************************************
 * SYSTEM TABLE SYNC
 ***************************************************************************/

/*************************************************************************
 * CUSTOM TABLE HANDLERS
 *************************************************************************/
 
/****************************************************************************
 * Function Name  : adslGroup_sync
 * Arguments      : 
 * Returns        : 
 * Variables      : 
 * Modifies       :  
 * Description    : sync dnmt status and ADSL group
 ****************************************************************************/
static int s_uprate=0;
static int found_table2=0;
void 
adslGroup_sync2 (unsigned int clientreg, void *clientarg)
{
  //  static time_t lastModemAntOamChange;
  int uprate, downrate;

  //    if(fileChanged(MODEM_ANT_STATUS_FILE, &lastModemAntOamChange))
  {
    //  adslChannelTable_sync();
  }
  getDownUpStreamRate(&downrate,&uprate);
  //DPRINTF("down %d up %d\n",downrate, uprate);
  s_uprate=uprate * 1000;
  adslChannelTable_sync2();
}

void adslChannelTable_sync2()
{
  int  ifIndex, auxValue; // foundRow=0;
  //float auxFloat=0;
  netsnmp_table_data_set *table_set = adslChannelTable_table_set2;
  netsnmp_table_row *row;
  //  netsnmp_table_data_set_storage *auxData;

  //  DEBUGMSGTL(("adsl", "adslChannelTable_sync create new row adslInterleave0 %d\n", ifIndex));
  //row = netsnmp_table_data_set_create_row_from_defaults(table_set->default_row);
  if( found_table2 == 0) {
    row = netsnmp_create_table_data_row();
    ifIndex = 8;
    netsnmp_table_row_add_index(row, ASN_INTEGER, (char *)&ifIndex, 4);
    //netsnmp_table_dataset_add_row(table_set, row);  /* add row to the table */
  } else {
    row = table_set->table->first_row;
  }
  ifIndex = 11;
  netsnmp_set_row_column(row, 1, ASN_INTEGER, (char *)(&ifIndex), 4); // 1
  //auxValue = ADSLLINETYPE_INTERLEAVEDONLY;
  //auxValue = 8064*1024; // bits
  //netsnmp_set_row_column(row, 2, ASN_INTEGER, (char *)(&auxValue), 4); // 2
  //auxValue = 12;
  netsnmp_set_row_column(row, 2, ASN_INTEGER, (char *)(&s_uprate), 4); // 2
  //auxValue = modem_ant_get_InlUpStreamSpeed(0);
  auxValue = 13;
  netsnmp_set_row_column(row, 3, ASN_INTEGER, (char *)(&auxValue), 4); // 3
  //auxValue = modem_ant_get_InlDnStreamSpeed(0);
  auxValue = 14;
  netsnmp_set_row_column(row, 4, ASN_INTEGER, (char *)(&auxValue), 4); // 4
  if( found_table2 == 0) {
    netsnmp_table_dataset_add_row(table_set, row);  /* add row to the table */
    found_table2 = 1;
  }
}
 
#endif 
