/*
 *  Interrupt service routines for Trendchip board
 */
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mipsregs.h>
#include <asm/tc3162/tc3162.h>

extern asmlinkage void tc3162_handle_int(void);
extern irq_desc_t irq_desc[NR_IRQS];

#define ALLINTS (IE_SW0 | IE_SW1 | IE_IRQ0 | IE_IRQ1 | IE_IRQ2 | IE_IRQ3 | IE_IRQ4 | IE_IRQ5)

static inline void tc3162_unmask_irq(unsigned int irq)
{
	VPint(CR_INTC_IMR) |=  (1 << irq);
}

static inline void tc3162_mask_irq(unsigned int irq)
{
	VPint(CR_INTC_IMR) &= ~(1 << irq);
}

static void tc3162_enable_irq(unsigned int irq)
{
	unsigned long flags;

  	save_and_cli(flags);
  	tc3162_unmask_irq(irq);
  	restore_flags(flags);
}

static void tc3162_disable_irq(unsigned int irq)
{
	unsigned long flags;

	save_and_cli(flags);
  	tc3162_mask_irq(irq);
  	restore_flags(flags);
}

static unsigned int tc3162_startup_irq(unsigned int irq)
{
   	tc3162_enable_irq(irq);
   	return 0;
}

static void tc3162_mask_and_ack_irq(unsigned int irq)
{
    tc3162_disable_irq(irq);
}

static void tc3162_end_irq(unsigned int irq)
{
    if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS)))
		tc3162_enable_irq(irq);
}

#define tc3162_shutdown_irq		tc3162_disable_irq

static struct hw_interrupt_type tc3162_irq_type = {
	"TC3162",
	tc3162_startup_irq,
	tc3162_shutdown_irq,
	tc3162_enable_irq,
	tc3162_disable_irq,
	tc3162_mask_and_ack_irq,
	tc3162_end_irq,
	NULL
};

void __init tc3162_irq_setup(void)
{
	unsigned int i;

	/* Disable all hardware interrupts */
	change_cp0_status(ST0_IM, 0x0);

	/* Initialize IRQ action handlers */
	for (i = 0; i < NR_IRQS; i++) {
		irq_desc[i].status	= IRQ_DISABLED;
		irq_desc[i].action	= NULL;
		irq_desc[i].depth	= 1;
		irq_desc[i].handler	= &tc3162_irq_type;
	}

	/* Set up the external interrupt exception vector */
	set_except_vector(0, tc3162_handle_int);

	/* Enable all interrupts */
	change_cp0_status(ST0_IM, ALLINTS);
}

void __init init_IRQ(void)
{
	tc3162_irq_setup();
}

void irq_bad(struct pt_regs *regs)
{
	/* This should never happen */
	printk(" CAUSE register = 0x%08lx\n", regs->cp0_cause);
	printk("STATUS register = 0x%08lx\n", regs->cp0_status);
	printk("   EPC register = 0x%08lx\n", regs->cp0_epc);
	show_regs(regs);
	panic("Stray interrupt, spinning...");
}
