#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/delay.h>

#include <asm/pci_channel.h>
#include <asm/tc3162/interrupts.h>
#include <asm/tc3162/tc3162_pci.h>

// Sam add for TC3162P2
#include <asm/tc3162/tc3162.h>
// End Sam add

#include <asm/tc3162/gpio.h>

#ifdef DEBUG
#define PRINTK printk
#else
#define PRINTK(arg...) NULL
#endif

int tc_processor=0;  // TC3162P = 1, TC3162L = 0

static struct resource pci_io_resource = {
	"pci IO space", 
	TC3162_PCI_DEVICE_IO_START,
	TC3162_PCI_DEVICE_IO_END,
	IORESOURCE_IO  // define in include/linux/ioport.h
};

static struct resource pci_mem_resource = {
	"pci memory space", 
	TC3162_PCI_DEVICE_MEM_START,
	TC3162_PCI_DEVICE_MEM_END,
	IORESOURCE_MEM  //define in include/linux/ioport.h
};

extern struct pci_ops tc3162_pci_ops;

struct pci_channel mips_pci_channels[] = {
		{&tc3162_pci_ops, &pci_io_resource, &pci_mem_resource, (int)NULL, (int)NULL },
		{ NULL,           NULL,             NULL,              (int)NULL, (int)NULL },
};

/*----------------------------------------------------------*/
int access_pci_config(struct pci_dev *dev, int where, u32 *data, int write)
{
	int slot = (dev->devfn >> 3) & 0x1F;
	int func =  dev->devfn & 0x7;

	if ( slot > 2 || func > 7 ) {
		*data = 0xFFFFFFFF;
		return -1;
	}

	if (slot != 0) // host-0x80000000, slot1-0x80000800, slot2-0x80001000
    where |= TC3162_PCI_HOST_CONFIG_MEM_ADDR |(1L << (10+slot)) | (func << 8);
	else
    where |= TC3162_PCI_HOST_CONFIG_MEM_ADDR | (func << 8);

// Sam add debug
//	printk("Sam add pci where is %08x, slot=%x:%x\n",where, slot, func);
// End Sam 
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_ADDR) = where & ~3;

	if ( write )
	 *(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA) = *data;
	else
	{
            *data = *(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA);
	}
	return PCIBIOS_SUCCESSFUL;
}

#define read_pci_config(dev, where, data)  access_pci_config(dev, where, data, 0)
#define write_pci_config(dev, where, data)  access_pci_config(dev, where, data, 1)

static int read_config_byte(struct pci_dev *dev, int where, u8 *val)
{
	u32 data;
    int ret = read_pci_config(dev, where, &data);
	int shift = (where&3) << 3;
    *val = (data >> shift) & 0x000000FF;
//printk("read_config_byte value=%08x\n\n",*val);
	return ret; 
}

static int read_config_word(struct pci_dev *dev, int where, u16 *val)
{
	u32 data;
    int ret = read_pci_config(dev, where, &data);
	int shift = (where&3) << 3;
    *val = (data >> shift) & 0x0000FFFF;
//printk("read_config_word value=%08x\n\n",*val);
	return ret; 
}

static int read_config_dword(struct pci_dev *dev, int where, u32 *val)
{
   int ret;
 //   return read_pci_config(dev, where, val);
       ret = read_pci_config(dev, where, val);
//printk("read_config_dword value=%08x\n\n",*val);
return ret;
}


static int write_config_byte(struct pci_dev *dev, int where, u8 val)
{
	u32 data;
	int shift = (where&3) << 3;
	if ( read_pci_config(dev, where, &data) )
	    return -1;
	data &= ~(0xFF << shift);
    data |= val << shift;
	return write_pci_config(dev, where, &data);
}

static int write_config_word(struct pci_dev *dev, int where, u16 val)
{
	u32 data;
	int shift = (where&3) << 3;
	if ( read_pci_config(dev, where, &data) )
	    return -1;
	data &= ~(0xFFFF << shift);
	data |= val << shift;
	return write_pci_config(dev, where, &data);
}

static int write_config_dword(struct pci_dev *dev, int where, u32 val)
{
    return write_pci_config(dev, where, &val);
}


struct pci_ops tc3162_pci_ops = {
    read_config_byte,
    read_config_word,
    read_config_dword,
    write_config_byte,
    write_config_word,
    write_config_dword
};


//--------------------------------------------------------------------
// needed by arch/mips/kernel/pci.c:pcibios_fixups[] needed by quirks.c:pci_do_fixups()
// TC3162 have two slots + 1 base
#define MAX_PCI_SLOT 3
static unsigned long io_start = TC3162_PCI_DEVICE_IO_START;
static unsigned long mem_start[MAX_PCI_SLOT+1]= \
		{TC3162_PCI_DEVICE_MEM_START, TC3162_PCI_DEVICE1_MEM_START, TC3162_PCI_DEVICE2_MEM_START, TC3162_PCI_UNSUPPORT_MEM_START};

void __init pcibios_fixup_resources(struct pci_dev *dev)   
{
    unsigned int slot, func, i, reg, phy_addr, size;
    struct resource *res;
    unsigned int j;
    u32 class;

    if ( dev->hdr_type != PCI_HEADER_TYPE_NORMAL )  return;

    slot = PCI_SLOT(dev->devfn);

    func = PCI_FUNC(dev->devfn);

    res  = &dev->resource[0];

    if (slot > MAX_PCI_SLOT || slot < 0) 
    {
	printk("no such PCI slot ..\n");
	return;
    }

#if 1    /* don't support; so mapped to ununsed address */
    pci_read_config_dword(dev, PCI_CLASS_REVISION, &class);
    class >>= 8;                                /* upper 3 bytes */
    dev->class = class;
    class >>= 8;
    if ((slot==2) && (class == PCI_CLASS_NETWORK_ETHERNET))
        slot = 3;
#endif
    for ( i = 0 ; i < 6 ; i++, res++ )  {
        if ( pci_resource_len(dev, i) == 0 )
            continue;

        size = res->end - res->start;

   	reg = PCI_BASE_ADDRESS_0 + (i << 2);

        pci_read_config_dword(dev, reg, &phy_addr);
PRINTK("????? READ default phy_addr=%x, res=%x:%x:%x:%x\n", phy_addr, res, res->start, res->end, size);

        switch ( phy_addr & PCI_BASE_ADDRESS_SPACE )  {

          case PCI_BASE_ADDRESS_SPACE_IO:
PRINTK("********* ADDRESS_SPACE_IO: 08x\n",io_start);
            if ( (io_start & size) != 0 )  // start address doesn't aligned the mutiple of size
                io_start = (io_start & ~size) + (size+1);
            phy_addr = io_start;
            io_start += size < 0x1000 ? 0x1000 : size+1;
            break;

          case PCI_BASE_ADDRESS_SPACE_MEMORY:
PRINTK("Sam tc3162_pci.c 180 ADDRESS_SPACE_MEMORY %08x\n",mem_start[slot]);
	    if (slot == 0)
		break;
	    //slot !=0
            if ( (mem_start[slot] & size) != 0 )  // start address doesn't aligned the mutiple of size
                mem_start[slot] = (mem_start[slot] & ~size) + (size+1);
            phy_addr = mem_start[slot];
            mem_start[slot] += size < 0x1000 ? 0x1000 : size+1;
            break;

           default:
PRINTK("????????unknow type %x\n", phy_addr & PCI_BASE_ADDRESS_SPACE);
            continue;
        }

        printk( "**PCI: slot %d, func %d, base %d remap to %08x (size=%x) (reg=%08x)\n", slot, func, i, phy_addr, size, reg );

        pci_write_config_dword(dev, reg, phy_addr);

        res->start = phy_addr;

        res->end   = phy_addr + size;
    }
}


// needed by arch/mips/kernel/pci.c: pcibios_init()
void __init pcibios_fixup(void)
{
}

// needed by arch/mips/kernel/pci.c: pcibios_init()
void __init pcibios_fixup_irqs(void)
{
	unsigned int slot, func;
	unsigned char pin;
	struct pci_dev *dev;

	pci_for_each_dev(dev) {
		if (dev->bus->number != 0)
		    return;
		read_config_byte(dev, PCI_INTERRUPT_PIN, &pin);
		slot = PCI_SLOT(dev->devfn);
		func = PCI_FUNC(dev->devfn);
                if( slot == 1) 
		{      
			dev->irq = INTR_PCI_INTA;
		}
		else if (slot == 2 )
		{
                	dev->irq = INTR_PCI_INTB;
                }
		//else 
                //   	printk("Unknown PCI slot number is %d\n",slot);

		write_config_byte(dev, PCI_INTERRUPT_LINE, dev->irq);
		//PRINTK(KERN_INFO "PCI: slot %d func %d irq %d ", slot, func, dev->irq);
		//PRINTK( "start=%X end=%X\n", dev->resource[0].start, dev->resource[0].end);
	}
}

// needed by drivers/pci/pci.c: pci_scan_bridge()
unsigned int pcibios_assign_all_busses(void)
{
	return 1;
}


//---------------------------------------------------------
int tc3162_pci_init(void)
{
   unsigned long pci_reg_data;
   unsigned short gpiodata;

   printk("============================================================\n");
   printk("TC3162 PCI interface initialize\n");

   // reset the PCI bus
    *(volatile unsigned int *)(0xbfb00088) &= ~(1<<31);
    mdelay(100);
    *(volatile unsigned int *)(0xbfb00088) = (1<<31);
    mdelay(500);
	
   // Sam add for read PCI register to determine TC3162P or TC3162L
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_ADDR) = (TC3162_PCI_HOST_CONFIG_MEM_ADDR | PCI_VENDOR_ID);
   pci_reg_data = *(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA);
   if (pci_reg_data == 0x31629005) // yes,we get correct TC3162P PCI VENDOR_ID and DEVICE_ID, this is TC3162P
   {
	tc_processor = 1; //tc_processor = TC3162P
      
   // just assign tc3162 pci host memory base address0 any value
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_ADDR) = (TC3162_PCI_HOST_CONFIG_MEM_ADDR | PCI_BASE_ADDRESS_0);
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA) = 0x0;

   // write cache line size (0x08 -- 4 DWORD of pre-fetched data ) and
   // latency timer (0x20) , these value are suggested by trendchip 
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_ADDR) = (TC3162_PCI_HOST_CONFIG_MEM_ADDR | PCI_CACHE_LINE_SIZE);
   //*(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA) = 0x2008;
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA) = 0x2008;

   /* write pci_command (0x0016) and pci_status (0xf900), these values are suggested by trendchip

      pci_command = PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER | PCI_COMMAND_INVALIDATE
      
      write 1 to these pci_status register clear these status bit 
      pci_status  = PCI_STATUS_DETECTED_PARITY | PCI_STATUS_SIG_SYSTEM_ERROR | PCI_STATUS_REC_MASTER_ABORT |
		    PCI_STATUS_REC_TARGET_ABORT |  PCI_STATUS_SIG_TARGET_ABORT |  PCI_STATUS_PARITY
   */
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_ADDR) = (TC3162_PCI_HOST_CONFIG_MEM_ADDR | PCI_COMMAND);
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA) = 0xf9000016;

    
   // Enable slot1 bus master and memory space access
   // TC3162_PCI_SLOT1_CONFIG_MEM_ADDR
	#if 1 
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_ADDR) = (TC3162_PCI_SLOT1_CONFIG_MEM_ADDR | PCI_COMMAND);
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA) = 0x06;
	#endif

   // write slot1 cache line size (0x00) and latency timer (0x40) for RaLink wireless driveri or
   // write slot1 cache line size (0x00) and latency timer (0x28) for Inprocomm wireless driver
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_ADDR) = (TC3162_PCI_SLOT1_CONFIG_MEM_ADDR | PCI_CACHE_LINE_SIZE);

   // Ralink  
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA) = 0x00004000;

   // TC3162_PCI_SLOT2_CONFIG_MEM_ADDR
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_ADDR) = (TC3162_PCI_SLOT2_CONFIG_MEM_ADDR | PCI_COMMAND);
   *(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA) = 0x06;

   *(volatile unsigned long *)(TC3162_PCI_CONFIG_ADDR) = (TC3162_PCI_SLOT2_CONFIG_MEM_ADDR | PCI_CACHE_LINE_SIZE);

   *(volatile unsigned long *)(TC3162_PCI_CONFIG_DATA) = 0x00004000;

   return 0;
   }
   else {// We get incorrect PCI VENDOR_ID and DEVICE_ID, this is TC3162L
	return -1;
   }
}


