/*  
 * Miscellaneous definitions used to initialise the interrupt vector table
 * with the machine-specific interrupt routines.
 *
 * Written by Robert Wu, Billion Co.
 * Support for TrendChip TC3162
 */

#ifndef __TC3162_INTERRUPTS_H__ 
#define __TC3162_INTERRUPTS_H__ 

/*
 * TC3162 Interrupts
 */
#include <asm/addrspace.h>

/*
 * Register mapping 
 */
#define TC3162_INTR_BASE	(KSEG1 + 0x1FB40000)
#define TC3162_ITR_REG		(TC3162_INTR_BASE + 0x0)
#define TC3162_IMR_REG		(TC3162_INTR_BASE + 0x4)
#define TC3162_IPR_REG		(TC3162_INTR_BASE + 0x8)
#define TC3162_ISR_REG		(TC3162_INTR_BASE + 0xC)
#define TC3162_IPSR0_REG	(TC3162_INTR_BASE + 0x10)
#define TC3162_IPSR1_REG	(TC3162_INTR_BASE + 0x14)
#define TC3162_IPSR2_REG	(TC3162_INTR_BASE + 0x18)
#define TC3162_IPSR3_REG	(TC3162_INTR_BASE + 0x1C)
#define TC3162_IPSR4_REG	(TC3162_INTR_BASE + 0x20)
#define TC3162_IPSR5_REG	(TC3162_INTR_BASE + 0x24)
#define TC3162_IPSR6_REG	(TC3162_INTR_BASE + 0x28)
#define TC3162_IPSR7_REG	(TC3162_INTR_BASE + 0x2C)
#define TC3162_IntVec_REG	(TC3162_INTR_BASE + 0x30)

/*
 *  Interrupt Controller Source Connection
 *
 */
#define INTR_MIN	0
#define INTR_UART	0
#define INTR_TIMER0	4
#define INTR_TIMER1	5
#define INTR_TIMER2	6
#define INTR_TIMER3	7
#define INTR_TIMER4	8
#define INTR_Watchdog	9
#define INTR_GPIO	10
#define INTR_APB_DMA_C0 14
#define INTR_APB_DMA_C1 15
#define INTR_AFE_Gasp	18
#define INTR_DMT	19
#define INTR_AHB_Bus_Er 20
#define INTR_EMAC	21
#define INTR_ATM_SAR	22
#define INTR_USB	23
#define INTR_PCI_INTA	24
#define INTR_PCI_INTB	25
#define INTR_PCI_Bus_Er	26
#define INTR_SOFT_Int	31
#define INTR_MAX	INTR_SOFT_Int	/* Maximum IRQ */

#define NR_INTS		(INTR_MAX + 1)

#ifndef __ASSEMBLY__

/*
 * Interrupt table structure to hide differences between different
 * systems such.
 */
extern long cpu_mask_tbl[NR_INTS];
extern long cpu_irq_nr[NR_INTS];

/*
 * Common interrupt routine prototypes
 */
extern void tc3162_intr_unimplemented(void);

#endif /* !__ASSEMBLY__ */
#endif /* __TC3162_INTERRUPTS_H__  */

