/*
** $Id: privacy.c 1.36 2004/10/04 17:17:39 cmchang Exp $
*/

/*************************************************************************
* Copyright (c) 2002-2003 Inprocomm, Inc.
*
* All rights reserved. Copying, compilation, modification, distribution
* or any other use whatsoever of this material is strictly prohibited
* except in accordance with a Software License Agreement with
* Inprocomm, Inc.
**************************************************************************/
/*************************************************************************
GENERAL DESCRIPTION
********************
This file contains the privacy-related routines of Windows driver for
INPROCOMM IPN2220 PCI 802.11 Wireless LAN Adapter.
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/
#define CRC32_POLY      0xedb88320     /* AUTODIN II, Ethernet, & FDDI */

/* The scramble key used to scramble/descramble encryption keys. */
#define SCRAMBLE_KEY    {0x49, 0x76, 0x61, 0x6E, 0x20, 0x6C, 0x6F, 0x76, \
                         0x65, 0x20, 0x41, 0x6D, 0x79, 0x5E, 0x5F, 0x5E}

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*_________________________________________________________________________
**  privacyInitialize
**
**  descriptions: This routine is called to initialize the privacy-related
**                parameters.
**  parameters:
**              adapter_p - Pointer to the adapter structure
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
privacyInitialize(IN PMP_ADAPTER adapter_p)
{
    adapter_p->swKeyTableSize = IPN_KEY_TABLE_SIZE;

    /* Init H/W key table */
    pauClearHwKeyTable(adapter_p);

    /* Init S/W key table */
    privacyClearSwKeyTable(adapter_p);

}

/*_________________________________________________________________________
**  privacyClearSwKeyTable
**
**  description: This function clear the S/W key table
**  parameters:  adapter_p - Pointer to the adapter object data area.
**  return:
**  note:
**_________________________________________________________________________*/
VOID
privacyClearSwKeyTable(IN PMP_ADAPTER adapter_p)
{
    /* Clean S/W key table */
    NdisZeroMemory(&adapter_p->swKeyTable[0],
        sizeof(IPN_SW_KEY_TABLE_ENTRY) * IPN_KEY_TABLE_SIZE);
}

/*_________________________________________________________________________
**  privacySeekForEntry
**
**  description:
**  parameters:  adapter_p - Pointer to the adapter object data area.
**
**  return:      TRUE, specific key is found
**               FALSE, specific key is not found
**  note:
**_________________________________________________________________________*/
BOOLEAN
privacySeekForEntry(IN PMP_ADAPTER priv_p,
                    IN IPN_KEY_STRUC *key_p,
                    OUT PULONG keyIndex_p)
{
        IPN_KEY_STRUC   *keyEntry_p;
        int             i;

        /* Search old and same entry */
        for (i = 0; i < priv_p->swKeyTableSize; i++) {

                if (priv_p->swKeyTable[i].entryUsed) {
                        keyEntry_p = &priv_p->swKeyTable[i].key;

                        if (keyEntry_p->chkKeyId == key_p->chkKeyId &&
                            keyEntry_p->chkAddr1 == key_p->chkAddr1 &&
                            keyEntry_p->chkAddr2 == key_p->chkAddr2) {

                                if (keyEntry_p->chkKeyId &&
                                    keyEntry_p->keyId != key_p->keyId) {
                                        continue;
                                }
                                if ((keyEntry_p->chkAddr1 || keyEntry_p->chkAddr2) &&
                                    ipn_memcmp(keyEntry_p->addr, key_p->addr, 6)) {
                                        continue;
                                }
                                *keyIndex_p = (ULONG) i; /* Old one exists */
                                return TRUE;
                        }
                } /* end of entryUsed */
        }

        return FALSE; /* Not hit */
}

/*_________________________________________________________________________
**  privacySeekForEmptyEntry
**
**  description: This function seeks empty key entry from the S/W key table
**  parameters:  adapter_p - Pointer to the adapter object data area.
**               keyType - 0, Group key,
**                         1, Pairwise key,
**                         2, Multicast/Broadcast key
**               keyEntry_p - Pointer to put the output key entry index in
**                            S/W key table
**  return:      TRUE, empty key entry is found
**               FALSE, no empty key entry available
**  note:
**_________________________________________________________________________*/
BOOLEAN
privacySeekForEmptyEntry(IN PMP_ADAPTER priv_p,
                         IN BOOLEAN isGroupKey,
                         OUT PULONG keyIndex_p)
{
        int                     i, index;

        if (isGroupKey) {
                /* Search new and free entry, larger than current ones */
                index = -1;
                for (i = 0; i < priv_p->swKeyTableSize; i++) {
                        if (priv_p->swKeyTable[i].entryUsed) {
                                index = -1; /* <0 */
                        }
                        else if (index < 0) {
                                index = i;
                        }
                }

                if (index >= 0) {
                        *keyIndex_p = (ULONG) index;
                        return TRUE;
                }
        }
        else {
                /* Search new and free entry, as small as possible */
                for (i = 0; i < priv_p->swKeyTableSize; i++) {
                        if (!priv_p->swKeyTable[i].entryUsed) {
                                *keyIndex_p = (ULONG) i;
                                return TRUE;
                        }
                }
        }

        return FALSE; /* No adaptive free entry */
}


