/*
** $Id: sdl_Dist_serv.c 1.87 2004/09/23 19:07:02 cmchang Exp $
*/

/*! \file sdl_Dist_serv.c
    \brief IEEE 802.11 standard MAC SDL, BLOCK Distribution_Service

    This block interfaces between the AP function and the Distribution
    System Medium, hence is only present at APs.
    In order to permit an LLC entity colocated with an AP to communicate via
    both the WM and the DSM, MAC_Data_Service at the AP interacts with this
    block. This causes frames originating at the station containing the AP to be
    treated equivalently to frames originating at any of the other stations
    associated with that AP.
*/

/******************************************************************************
Copyright (c) 2002-2003 Integrated Programmable Communications, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Integrated Programmable Communications, Inc.
******************************************************************************/

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"
/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

#define FORWARD_TO_ALL          0       /* group address */
#define FORWARD_TO_ETHER        1
#define FORWARD_TO_WLAN         2
#define FORWARD_DISCARD         3

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/


/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param  priv_p       pointer to i2220mgmt_t, the device's private structure
* \param  da
* \param  sta_pp
* \return
*/
/*----------------------------------------------------------------*/
int distribute_decision(i2220mgmt_t *priv_p, MacAddr_t da,
                        sta_info_t *send_sta_p, sta_info_t **sta_pp)
{
        sta_info_t      *sta_p;

        KDBG_ENTER;
        if (sdl_isGroup(da)) {
                *sta_pp = NULL;
                if ((priv_p->relay_mode == RelayMode_DATA_PIPE) ||
                    (priv_p->relay_mode == RelayMode_SWITCH_NO_FORWARD)) {
                        KDBG_LOG_RX(KDBG_DEBUG, "Forward to ETHER\n");
                        KDBG_EXIT;
                        return FORWARD_TO_ETHER;
                }
                else {
                        KDBG_LOG_RX(KDBG_DEBUG, "Forward to WLAN/ETHER\n");
                        KDBG_EXIT;
                        return FORWARD_TO_ALL;
                }
        }

        sta_p = task_sta_lookup(priv_p, da);
        if (sta_p && sta_p->aid > 0){
                *sta_pp = sta_p;

                if (priv_p->relay_mode == RelayMode_SWITCH_NO_FORWARD) {
                        KDBG_LOG_RX(KDBG_DEBUG, "Forward discard\n");
                        KDBG_EXIT;
                        return FORWARD_DISCARD;
                }
                else if (priv_p->relay_mode == RelayMode_DATA_PIPE) {
                        return FORWARD_TO_ETHER;
                }
                else {
                #if SUPPORT_MSSID
                /*
                **  We invoke checking VLAN_ID when in WLAN Swtich mode.
                **  If either SSID enables VLAN TAG, and their VLAN TAGs are not
                **  equal, then we forward this packet upward. Otherwise, AP
                **  driver forward this packet to the destination station.
                */
                        ssid_info_t *dest_p;
                        ssid_info_t *send_p;

                        dest_p = &priv_p->ap.ssid_list[sta_p->whichSsid];
                        send_p = &priv_p->ap.ssid_list[send_sta_p->whichSsid];
                        if (dest_p->vlan_tag_enabled != send_p->vlan_tag_enabled
                            || (send_p->vlan_tag_enabled &&
                                dest_p->vlanid != send_p->vlanid)) {

                                KDBG_LOG_RX(KDBG_DEBUG, "Forward to ETHER\n");
                                return FORWARD_TO_ETHER;
                        }
                #endif
                        KDBG_LOG_RX(KDBG_DEBUG, "Forward to WLAN\n");
                        KDBG_EXIT;
                        return FORWARD_TO_WLAN;
                }
        }

        KDBG_LOG_RX(KDBG_DEBUG, "Forward to ETHER\n");
        KDBG_EXIT;
        return FORWARD_TO_ETHER;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param  priv_p       pointer to i2220mgmt_t, the device's private structure
* \param  pdu
* \param  sta_p
* \return
*/
/*----------------------------------------------------------------*/
KAL_IRAM int rx_do_forward(i2220mgmt_t *priv_p, Frame_t *pdu,
                        sta_info_t *send_sta_p)
{
        MacAddr_t               da;
        kal_skb_t               *new_skb;
        sta_info_t              *sta_p;
        int                     txPkt_len;
//printk("Sam 169 rx_do_forward\n");
        /* Decide DA address only by toDs bit */
        da = sdl_toDs(pdu) ? pdu->addr3 : pdu->addr1;

        /* sta_p is the new wireless destination STA record */
        switch (distribute_decision(priv_p, da, send_sta_p, &sta_p)) {
        case FORWARD_TO_ALL:
                /* bcast frame is not forwarded to WDS */
                txPkt_len = sizeof(Frame_t) + pdu->LLClength;
                if ((new_skb = kal_alloc_skb(txPkt_len)) != NULL) {
                        Frame_t *tpdu;

                        /* The starting address should be 4-wise boundary */
                        ASSERT((((UINT32)kal_skb_data(new_skb))&3)==0);

                        kal_skb_put(new_skb, txPkt_len);
                        tpdu = (Frame_t *)kal_skb_data(new_skb);
                        tpdu->skb = new_skb;
                        tpdu->mac_ctl = MACCTL_SDU_FRAME_IN_SKB |
                                MACCTL_CHECK_SLEEP | MACCTL_CHECK_CIPHER;
                        tpdu->LLCdata = (UINT8*)tpdu + sizeof(Frame_t);
                        tpdu->LLClength = pdu->LLClength;

                        wlanhdr_to_wlanhdr(priv_p, tpdu, pdu);
                        ipn_memcpy(tpdu->LLCdata, pdu->LLCdata, pdu->LLClength);

                        sdu_send_request(priv_p, tpdu, sta_p, SYN_contention);
                }
		else printk("FORWARD_TO_ALL: kal_alloc_skb fail\n");
        case FORWARD_TO_ETHER:
//printk("Sam rx_do_forword 198 FORWARD_TO_ETHER\n");
                wlanhdr_to_ether(priv_p->netdev, pdu);
        #if SUPPORT_MSSID
                append_vlang_tag(pdu->skb,&priv_p->ap.ssid_list[send_sta_p->whichSsid]);
        #endif
                kal_netdev_rx(pdu->skb, priv_p->netdev);
                break;
        case FORWARD_TO_WLAN:
                wlanhdr_to_wlanhdr(priv_p, pdu, pdu);
                pdu->mac_ctl |= MACCTL_CHECK_SLEEP | MACCTL_CHECK_CIPHER;
                sdu_send_request(priv_p, pdu, sta_p, SYN_contention);
                break;
        case FORWARD_DISCARD:
        default:
                KDBG_EXIT;
                return RX_DRV_HANDLED_PACKET;
        }

        KDBG_EXIT;
        return RX_INDICATE_PACKET_UPWARD;
}

#if SUPPORT_TKIP
/*----------------------------------------------------------------*/
/*!
* \brief  Check secMode to calculate TKIP-MIC
*
* \param  priv_p       pointer to i2220mgmt_t, the device's private structure
* \param  pdu
* \param  sta_p
* \param  rfb_p
* \return
*/
/*----------------------------------------------------------------*/
int do_chk_tkipMIC(i2220mgmt_t *priv_p, Frame_t *pdu, sta_info_t *sta_p,
                UINT keyId, UINT8 *key_p)
{
        FragSdu_t       *fsdu;

        KDBG_ENTER;
        if ((fsdu = mk_fsdu(priv_p, sta_p, pdu)) == NULL) {
                sdlu_release_sdu(pdu);
                KDBG_LOG_RX(KDBG_WARN, "Not enough memory for mk_fsdu\n");
                KDBG_EXIT;
                return -1;
        }
        fsdu->cipherSuite = CIPHER_SUITE_TKIP;
        fsdu->encryptionKeyId = (UINT8)keyId;
        ipn_memcpy(fsdu->encryptKey, key_p, 32);

        /* Proceed to generate MIC to check counterMeasure */
        Decide_Sa_Da(pdu, &fsdu->sa_p, &fsdu->da_p);
        tkipHwGenMic(priv_p, fsdu, SECQ_RX, TRUE);

        KDBG_EXIT;
        return 0;
}
#endif /* SUPPORT_TKIP */

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param  priv_p       pointer to i2220mgmt_t, the device's private structure
* \param  pdu
* \param  sta_p
* \param  rfb_p
* \return
*/
/*----------------------------------------------------------------*/
int rx_process_data(i2220mgmt_t *priv_p, Frame_t *pdu,
                                sta_info_t *sta_p, rfb_t *rfb_p)
{
#if SUPPORT_TKIP
        UINT            keyId;
        UINT8           *key_p;
#endif
//printk("Sam 275 rx_process_data\n");
        KDBG_ENTER;
        if (!sta_p || sta_p->aid == 0) {
                /* not associated yet, send back deauth frame */
                sdlf_send_deauth(priv_p, pdu->addr2, SYN_class3_err);
                KDBG_EXIT;
                return RX_DRV_HANDLED_PACKET;
        }

        if (sdl_ftype(pdu) == SYN_null_frame) {
                /* NULL data, just ignore now */
                KDBG_LOG_RX(KDBG_DEBUG, "Rx NULL frame from "
                        MACSTR "\n", MAC2STR(pdu->addr2));
                KDBG_EXIT;
                return RX_DRV_HANDLED_PACKET;
        }

        /* Check if the traffic should be blocked */
        if (priv_p->rsnBlockingTraffic ||sta_p->rsnBlockingTraffic ||
            (priv_p->BCcipherSuite && !sdl_wepBit(pdu))) {

                if (priv_p->rsnBlockingTraffic ||
                    !CheckFor1xFrame(pdu->LLCdata, pdu->LLClength)) {
                        KDBG_LOG_RX(KDBG_DEBUG, "Throw off the packet from "
                                MACSTR "\n", MAC2STR(pdu->addr2));
                        KDBG_EXIT;
                        return RX_DRV_HANDLED_PACKET;
                }
        }

#if SUPPORT_TKIP
        /* Check if counterMeasure is needed */
        if (sdl_wepBit(pdu) && rfb_p->sw.secMode == CIPHER_SUITE_TKIP &&
                RxMsduDecrypt(priv_p, sta_p, pdu, &keyId, &key_p)) {

        #if WP_3210_WORKAROUND
                UINT8   *sa_p, *da_p;
                UINT8   tkipMIC[8];

                Decide_Sa_Da(pdu, &sa_p, &da_p);
                Run_Michael_Plus(&key_p[24], da_p, sa_p, pdu->LLCdata,
                                pdu->LLClength, tkipMIC);

                if (ipn_memcmp(tkipMIC, pdu->LLCdata + pdu->LLClength, 8)) {
                        kal_tkip_micFailure(priv_p->netdev, pdu->addr2);
                        return RX_DRV_HANDLED_PACKET;
                }

        #elif SUPPORT_SW_TKIP_MIC
                if (pdu->LLClength > priv_p->HWMICthreshold) {
                        /* Need to check TKIP-MIC of this msdu */
                        do_chk_tkipMIC(priv_p, pdu, sta_p, keyId, key_p);
                        return RX_INDICATE_PACKET_UPWARD;
                }
                else {
                        UINT8   *sa_p, *da_p;
                        UINT8   tkipMIC[8];

                        Decide_Sa_Da(pdu, &sa_p, &da_p);
                        tkipSwGenMic(&key_p[24], pdu->LLCdata, pdu->LLClength,
                                sa_p, da_p, 0, tkipMIC);

                        if (ipn_memcmp(tkipMIC, pdu->LLCdata + pdu->LLClength, 8)) {
                                kal_tkip_micFailure(priv_p->netdev, pdu->addr2);
                                return RX_DRV_HANDLED_PACKET;
                        }
                }
        #else
                /* Need to check TKIP-MIC of this msdu */
                do_chk_tkipMIC(priv_p, pdu, sta_p, keyId, key_p);
                return RX_INDICATE_PACKET_UPWARD;
        #endif
        }
#endif

        return rx_do_forward(priv_p, pdu, sta_p);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param  priv_p       pointer to i2220mgmt_t, the device's private structure
* \param  pdu
* \param  sta_p
* \return
*/
/*----------------------------------------------------------------*/
int wds_rx_do_forward(i2220mgmt_t *priv_p, Frame_t *pdu, sta_info_t *remote_ap_p)
{
#if !WLAN_INCLUDE_WDS
        return RX_DRV_HANDLED_PACKET;
#else
        kal_net_device_t *dev;
        wdslink_info_t* wds = NULL;
//printk("Sam 370 wds_rx_do_forward\n");
        if((wds=wds_by_mac(priv_p, pdu->addr2)) != NULL)
                dev = wds->dev;
        else{
                KDBG_LOG_RX(KDBG_ERR, "No this wds "MACSTR"\n",MAC2STR(pdu->addr2));
                return RX_DRV_HANDLED_PACKET;
        }

        /*Because IPN2220 has a bug before ECO, if we received a data frame which is QoS frame,
            we can assume it has this bug. And we must send CCMP with QoS to it.
        */
        if(sdl_dataQos(pdu)){
                wds->mode = 0;
        }

        wlanhdr_to_ether(dev, pdu);
        kal_netdev_rx(pdu->skb, dev);
        return RX_INDICATE_PACKET_UPWARD;
#endif
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param  priv_p       pointer to i2220mgmt_t, the device's private structure
* \param  pdu
* \param  remote_ap_p Representing remote Access Point
* \param  rfb_p
* \return
*/
/*----------------------------------------------------------------*/
int wds_rx_process_data(i2220mgmt_t *priv_p, Frame_t *pdu,
                                sta_info_t *remote_ap_p, rfb_t *rfb_p)
{
#if !WLAN_INCLUDE_WDS
        return RX_DRV_HANDLED_PACKET;
#else

#if SUPPORT_TKIP
        UINT            keyId;
        UINT8           *key_p;
#endif

        KDBG_ENTER;
        if ((!remote_ap_p) || (!remote_ap_p->wds_link)) {
                /* This packet came from unknown source,
                 * therefore the packet should be recycled.
                 */
                KDBG_LOG_RX(KDBG_DEBUG, "WDS from unknown "MACSTR"\n",MAC2STR(pdu->addr2));
                return RX_DRV_HANDLED_PACKET;
        }

        if (sdl_ftype(pdu) == SYN_null_frame) {
                /* NULL data, just ignore now */
                KDBG_LOG_RX(KDBG_DEBUG, "Rx NULL frame from "
                        MACSTR "\n", MAC2STR(pdu->addr2));
                KDBG_EXIT;
                return RX_DRV_HANDLED_PACKET;
        }

        /* Check if the traffic should be blocked */
        if (priv_p->rsnBlockingTraffic ||remote_ap_p->rsnBlockingTraffic ||
            (remote_ap_p->KeyMapOn && !sdl_wepBit(pdu))) {

                if (priv_p->rsnBlockingTraffic ||
                    !CheckFor1xFrame(pdu->LLCdata, pdu->LLClength)) {
                        KDBG_LOG_RX(KDBG_DEBUG, "Throw off the packet from "
                                MACSTR "\n", MAC2STR(pdu->addr2));
                        KDBG_EXIT;
                        return RX_DRV_HANDLED_PACKET;
                }
        }



#if SUPPORT_TKIP
        /* Check if counterMeasure is needed */
        if (sdl_wepBit(pdu) && rfb_p->sw.secMode == CIPHER_SUITE_TKIP &&
            RxMsduDecrypt(priv_p, remote_ap_p, pdu, &keyId, &key_p)) {

        #if WP_3210_WORKAROUND
                UINT8   *sa_p, *da_p;
                UINT8   tkipMIC[8];

                Decide_Sa_Da(pdu, &sa_p, &da_p);
                Run_Michael_Plus(&key_p[24], da_p, sa_p, pdu->LLCdata,
                                pdu->LLClength, tkipMIC);

                if (ipn_memcmp(tkipMIC, pdu->LLCdata + pdu->LLClength, 8)) {
                        kal_tkip_micFailure(priv_p->netdev, pdu->addr2);
                        return RX_DRV_HANDLED_PACKET;
                }

        #elif SUPPORT_SW_TKIP_MIC
                if (pdu->LLClength > priv_p->HWMICthreshold) {
                        /* Need to check TKIP-MIC of this msdu */
                        do_chk_tkipMIC(priv_p, pdu, remote_ap_p, keyId, key_p);
                        return RX_INDICATE_PACKET_UPWARD;
                }
                else {
                        UINT8   *sa_p, *da_p;
                        UINT8   tkipMIC[8];

                        Decide_Sa_Da(pdu, &sa_p, &da_p);
                        tkipSwGenMic(&key_p[24], pdu->LLCdata, pdu->LLClength,
                                sa_p, da_p, 0, tkipMIC);

                        if (ipn_memcmp(tkipMIC, pdu->LLCdata + pdu->LLClength, 8)) {
                                kal_tkip_micFailure(priv_p->netdev, pdu->addr2);
                                return RX_DRV_HANDLED_PACKET;
                        }
                }
        #else
                /* Need to check TKIP-MIC of this msdu */
                do_chk_tkipMIC(priv_p, pdu, remote_ap_p, keyId, key_p);
                return RX_INDICATE_PACKET_UPWARD;
        #endif
        }
#endif

        KDBG_EXIT;
        return wds_rx_do_forward(priv_p, pdu, remote_ap_p);

#endif /* end of not defined(WLAN_INCLUDE_WDS) */
}

