/*
** $Id: tkip.c 1.32 2004/10/04 17:20:20 cmchang Exp $
*/

/*************************************************************************
* Copyright (c) 2000-2003 Integrated Programmable Communications, Inc.
*
* All rights reserved. Copying, compilation, modification, distribution
* or any other use whatsoever of this material is strictly prohibited
* except in accordance with a Software License Agreement with
* Inprocomm, Inc.
**************************************************************************/
/*************************************************************
************
GENERAL DESCRIPTION
********************
This implementation is developed based on IEEE 802.11i draft 2.3
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

void Decide_Sa_Da(Frame_t *macHeader_p, PUCHAR *sa_p, PUCHAR *da_p)
{
        KDBG_LOG_RX(KDBG_DEBUG,"Before TKIP MSDU Decapsulate:\n");

        /* Decide SA and DA according to ToDS and FromDS bits */
        if (!(sdl_toDs(macHeader_p)) &&
            (sdl_frDs(macHeader_p))) {
            *sa_p = macHeader_p->addr3;
            *da_p = macHeader_p->addr1;
        }
        else if ((sdl_toDs(macHeader_p)) &&
                 !(sdl_frDs(macHeader_p))) {
            *sa_p = macHeader_p->addr2;
            *da_p = macHeader_p->addr3;
        }
        else if (!(sdl_toDs(macHeader_p)) &&
                 !(sdl_frDs(macHeader_p))) {
            *sa_p = macHeader_p->addr2;
            *da_p = macHeader_p->addr1;
        }
        else if ((sdl_toDs(macHeader_p)) &&
                 (sdl_frDs(macHeader_p))) {
            *sa_p = macHeader_p->addr4;
            *da_p = macHeader_p->addr3;
        }
}

/*_________________________________________________________________________
** tkipHwGenMic
**
** description: TKIP Michael generation function.  This function is to generate
**              TKIP MIC by HW security engine.  The frame body for generate
**              MIC is linked in a TFCB with a single/multiple TBB(s). For
**              considering fragment need, the coalesce may be divided into
**              several fragments.  To padding necessary header/tailer for
**              generating MIC, the first fragment should reserved at least
**              16 bytes and the last fragment should reserved at least 24 bytes.
**              After completing, the MIC will be appended into the last fragment
**              and returned by an OUT parameter too.
** parameters:  adapter_p - Pointer to the adapter object data area.
**              swTfcb_p, SW TFCB pointer which links the whole frame body
**              mickey_p, pointer to MIC key
**              data_start_p, pointer to starting of the frame body
**              firstfragmentLen, the 1st fragment length, in byte(s)
**              sa_p, pointer to source address SA
**              da_p, pointer to destination address DA
**              priority, priority of IEEE 802.11 traffic class
**              data_end_p, pointer to end of the frame body
**              mic_p, pointer to 64-bit MIC
** return:
** note:
**__________________________________________________________________________*/
int tkipHwGenMic(i2220mgmt_t *priv_p, FragSdu_t *fsdu, TX_QUEUE_NUM qNo, bool chk_res)
{
        tdma_parameters_t   *tdma_p;
        tfcb_t              *tfcb_p, *tfcb_el_p, *tfcb_eol_p;
        tbb_t               *tbb_p;
        struct list_head    *lst;
        UINT8               *temp8_p, *mickey_p;
        UINT8               i, j, checksum, bytesPadding;
        UINT32              frameBodyLen = fsdu->sw.sdu->LLClength;
        UINT8               *frameBody_p = fsdu->sw.sdu->LLCdata;
        kal_dma_addr_t      dmaAddr;
#if SINGLE_SEC_QUEUE
        int                 oldq;
#endif

        KDBG_ENTER;

#if SINGLE_SEC_QUEUE
        oldq = qNo;
        qNo = SECQ_TX; /* Use Tx secQueue */
#endif

        tdma_p = &priv_p->tdma[qNo];
        if ((tfcb_eol_p = tdma_p->tail_p) == NULL) {
                /* it's firt time to startup this TDMA channel */
                get_chain_head(&priv_p->sec_tfcb_chain,tfcb_t,tdma_p->tail_p);
                if (tdma_p->tail_p == NULL) {
                        sdlu_release_fsdu(fsdu);
                        KDBG_LOG_TX(KDBG_WARN,"tdma %d is empty.\n", qNo);
                        KDBG_EXIT;
                        return TX_TFCB_FAILURE;
                }
                tdma_p->tail_p->hw.tfdCtrlFlag = TFCB_FLAG_EOL;
                tdma_p->head_p = tdma_p->tail_p;
                tdma_p->count++;
        }

        /* Calculate the total required TFCBs/TBBs */
        if (chk_res && ((!tx_check_TDMA(&priv_p->sec_tfcb_chain, 1, 0)) ||
            sdlQ_Length(&tdma_p->waitQ) != 0)) {
                /* No TFCB or TBB, put it into waitQ */
                sdlQ_Qlast_limit(&tdma_p->waitQ, fsdu, 100);
                KDBG_EXIT;
                return TX_TFCB_FAILURE;
        }

        tfcb_p = tdma_p->tail_p;

        /* Important: clear completion status and tfdCtrlFlag except EOL */
        tfcb_p->hw.tfdCtrlFlag &= TFCB_FLAG_EOL;
        tfcb_p->tkipMic[39] = 0;
        *(UINT32*)&tfcb_p->tkipMic[44] = 0; /* OK in whatever endian */

        /* Link the association between fsdu and tfcb_p */
        tfcb_p->sw.fsdu = fsdu;

        /*-----------------------------------------------------
         * MPDU header (0x48--0x6c) and security (0x10--0x28)
         * ==> including length fields
         * Note: for big endian, some short and long fields
         *       are swapped at the end
         *----------------------------------------------------*/
        /* Prepare TFCB */
        tfcb_p->hw.tfdCtrlFlag |= TFCB_FLAG_SEC;

        ASSERT((offsetof(tfcb_t, tkipMic[0]) & 3) ==0);
        tfcb_p->hw.secModeDestAddr = host2chip_32(
                        tfcb_p->dmaAddr + offsetof(tfcb_t, tkipMic[0]));
        tfcb_p->hw.totalTxDataLen = frameBodyLen + 16; /* DA + SA+ Priority + 3 bytes of 0 + data length*/

#if SINGLE_SEC_QUEUE
        if (oldq == SECQ_RX) {
                tfcb_p->hw.extendedIv = 0xa0a0a0a0;
                mickey_p = &fsdu->encryptKey[24];
        }
        else {
                tfcb_p->hw.extendedIv = 0xa1a1a1a1;
                mickey_p = &fsdu->encryptKey[16];
        }
#else
        tfcb_p->hw.extendedIv = 0xabababab; /* Non-zero value */
        mickey_p = (qNo == SECQ_TX) ?
                (&fsdu->encryptKey[16]) : (&fsdu->encryptKey[24]);
#endif

        /* Fill MIC key */
        for (i = 0; i < 8; i++) {
                tfcb_p->hw.encryptionKey[i] = *(mickey_p + i);
        }

        //For 3 tbb
        i = 0;
        list_for_each(lst,&tfcb_p->sw.tbb_list) {
                tbb_p = list_entry( lst, tbb_t, sw.link );
                i++;
                if (i == 1) {
                        /* Prepare 1st TBB (DA, SA, and Priotity) */
                        temp8_p = tbb_p->hw.revBuf;
                        /* padding DA, SA, Priority */
                        for (j = 0; j < 6; j++) {
                                * (temp8_p + j) = * (fsdu->da_p + j);
                                * (temp8_p + 6 +j) = * (fsdu->sa_p + j);
                        }
                        * (temp8_p + 12) = 0; /* Priority */
                        for (j = 13; j < 16;j++)
                                * (temp8_p + j) = 0;

                        dbg_tx_packet(priv_p,tbb_p,DBFLG_TX_TBB);
                }
                else if (i == 2) {
                        /* Prepare 2nd TBB (Payload) */
                        dmaAddr = kal_pci_map_single(priv_p->pcidev, frameBody_p,
                                                frameBodyLen, PCI_DMA_TODEVICE);
                        tbb_p->hw.txDataBufferPointer = host2chip_32(dmaAddr);
                        tbb_p->hw.txDataBufferLen = host2chip_16(frameBodyLen);
                        tbb_p->do_map_unmap = TRUE;

                        tbb_p->hw.tbdChecksum = 0;
                        nicCalculateChecksumByte((PVOID) tbb_p, TBB_TBD_SIZE, &checksum);
                        tbb_p->hw.tbdChecksum  = ~checksum;

                        dbg_tx_packet(priv_p,tbb_p,DBFLG_TX_TBB);
                }
                else if (i == 3) {
                        /* Prepare 3rd TBB (padding of 0x5a and 0x00) */
                        bytesPadding = 8 - (frameBodyLen & 3);

                        tbb_p->hw.txDataBufferLen = host2chip_16( bytesPadding);
                        tbb_p->hw.txDataBufferPointer =
                                host2chip_32(tbb_p->dmaAddr + offsetof(hw_tbb_t, revBuf));
                        tfcb_p->hw.totalTxDataLen += (UINT16) bytesPadding;
                        tbb_p->do_map_unmap = FALSE;

                        tbb_p->hw.tbdChecksum = 0;
                        nicCalculateChecksumByte((PVOID) tbb_p, TBB_TBD_SIZE, &checksum);
                        tbb_p->hw.tbdChecksum = ~checksum;

                        dbg_tx_packet(priv_p,tbb_p,DBFLG_TX_TBB);
                }
        }

        get_chain_head(&priv_p->sec_tfcb_chain, tfcb_t, tfcb_el_p);
        ASSERT(tfcb_el_p);
        tfcb_el_p->hw.tfdCtrlFlag = TFCB_FLAG_EOL;

        tfcb_p->nextTfcb = tfcb_el_p;
        tfcb_p->hw.nextTfcbPointer = host2chip_32(tfcb_el_p->dmaAddr);
        /* Calculate TFD checksum */
        tfcb_p->hw.tfdChecksum = 0;
        nicCalculateChecksumByte((PVOID)&tfcb_p->hw, TFCB_TFD_SIZE, &checksum);
        tfcb_p->hw.tfdChecksum = ~checksum;

#if __CPU_ENDIAN__ != __LITTLE_ENDIAN__
        tfcb_p->hw.totalTxDataLen = host2chip_16(tfcb_p->hw.totalTxDataLen);
#endif
        /* clear EOL (tfdCtrlFlag -= TFCB_FLAG_EOL) */
        tdma_p->tail_p->hw.tfdChecksum += TFCB_FLAG_EOL;
        tdma_p->tail_p->hw.tfdCtrlFlag &= ~TFCB_FLAG_EOL;

        tdma_p->count += 1;
        tdma_p->tail_p = tfcb_el_p;    /* new EOL TFCB */

#if SINGLE_SEC_QUEUE && SUPPORT_PERF_MEASURE
        if (oldq == SECQ_RX){
            if ( fsdu->sw.sdu->skb && (skb_get_Flag(fsdu->sw.sdu->skb) & SKB_TSTAMP_IN_USE)) {
                UINT32 t1,t2;
                t1 = skb_get_TimeStamp(fsdu->sw.sdu->skb);
                // this is the time when upper LLC passed down the packet.
                fsdu->sw.txProcTime1 = t1;//skb_get_TimeStamp(fsdu->sw.sdu->skb);
                NIC_MCR_READ(priv_p, MCR_LTTR0, &t2);
                fsdu->sw.txProcTime2 = t2;
            }
        } else {
            if (fsdu->sw.txProcTime1) {
                UINT t3;
                NIC_MCR_READ(priv_p, MCR_LTTR0, &t3);
                fsdu->sw.txProcTime2 = t3;
            }
        }
#endif

        if (tfcb_eol_p == NULL) {
                KDBG_LOG_TX(KDBG_DEBUG,"TDMA%d is started.\n", qNo);
                dmaStartSecQueue(priv_p, qNo, tdma_p->head_p->dmaAddr);
        }
        else {
                KDBG_LOG_TX(KDBG_DEBUG,"TDMA%d is resumed.\n", qNo);
                dmaResumeSecQueue(priv_p, qNo);
        }

        KDBG_EXIT;
        return 0;
}

/*_________________________________________________________________________
** mk_fsdu
**
** description: When the RX receive a TKIP data frame, we need to packet the
**              data frame with a fsdu, and sending the fsdu to the tkipHwGenMic
**              function to cauclate the Rx Mic.
** parameters:  adapter_p - Pointer to the adapter object data area.
**              sta_p - the sta address of received frame
**              sdu - the received data frame
** return:      a fsdu
** note:
**__________________________________________________________________________*/
FragSdu_t *mk_fsdu(i2220mgmt_t * priv_p,sta_info_t *sta_p, Frame_t *sdu)
{
        FragSdu_t       *fsdu;
        KDBG_ENTER;

        /*-----------------------------------------------------
         * Construct FragSdu_t object, Initialize FragSdu structure
         *-----------------------------------------------------*/
        fsdu = (FragSdu_t*)kal_mempool_alloc(sizeof(FragSdu_t) +
                                        sizeof(FragMpdu_t));
        if (!fsdu) {
                KDBG_LOG_SDL(KDBG_CRIT,"Insufficient system memory resource\n");
                goto L_exit;
        }
        ipn_memset(fsdu, 0, sizeof(FragSdu_t));

        fsdu->sw.sdu = sdu;
        fsdu->sw.sta_p = sta_p; /* Important */

L_exit:
        KDBG_EXIT;
        return fsdu;
}


/*_________________________________________________________________________
** tkipHwMpduEncapsulate
**
** description: this function encapsulate MPDU frame body according to
**              IEEE 802.11i TKIP sepcification.
** parameters:  adapter_p - Pointer to the adapter object data area.
**              macHeader_p, pointer to Tx frame MAC header
**              frameBody_p, pointer to input text
**              frameBodyLen, length of frame body (in bytes), exclude length of
**                            IV, ExtIV and ICV
**              key_p, pointer to key
**              keyId, key index
**              resultFrameBodyLen_p, pointer to put the result frame body
**                                    length. (note: result frame body length
**                                    is equal to data payload legth
**                                    plus IV, ExtIV and ICV length)
** return:      TKIP_ENCAPSULATE_SUCCESS, if the MPDU is encapsulatable, i.e.
**              the MPDU is encrypted using RC4 algorithm
** note:        1. The frame body is expended by 12-byte, 4-byte for IV, ExtIV
**                 and ICV.
**              2. The result frame body is in the format:
**                 IV + ExtIV + PDU + ICV
**__________________________________________________________________________*/
UCHAR
tkipHwMpduEncapsulate(IN PMP_ADAPTER adapter_p,
                      IN OUT PTFCB_STRUC  tfcb_p,
                      IN UCHAR keyId,
                      IN PUCHAR key_p)
{
    UCHAR   iv_KeyId;
    UINT    i;

    DBGPRINT(RSN, TRACE, ("tkipHwMpduEncapsulate().\n"));
    /* Increase TSC by 1 */
    INC_48_BITS(adapter_p->TSC);

    /* Generate IV with TSC(TSC1 & TSC0) */
    /* Seed0, RC4Key[0] = TSC1*/
    tfcb_p->u_sciiv.s.initialVector[0] = *(PUCHAR)&adapter_p->TSC[1];
    /* Seed1, (TSC1 | 0x20) & 0x7F */
    tfcb_p->u_sciiv.s.initialVector[1] = (*(PUCHAR)&adapter_p->TSC[1] | 0x20) & 0x7F;
    /* Seed2, RC4Key[1] = TSC0*/
    tfcb_p->u_sciiv.s.initialVector[2] = *(PUCHAR)&adapter_p->TSC[0];

    /* Generate ExtIV with TSC(TSC2 ~ TSC5), Fill in SCI Extented IV  */
//    tfcb_p->sciExtendedIv = *(PULONG)&adapter_p->TSC[2];
    ipn_memcpy(&tfcb_p->sciExtendedIv,&adapter_p->TSC[2],4);
    if (keyId < 4) { /* default key */
        /* add key ID and padding zeros to IV, in byte 3 */
        iv_KeyId= (UCHAR)keyId << 6;    /* Key ID(bit 6 & 7) */
    }
    else { /* mapping key */
        /* set key ID to 0 */
        iv_KeyId = 0; /* mapping key. so, set key ID to 0 */
    }
    tfcb_p->u_sciiv.s.keyId = iv_KeyId;
    tfcb_p->u_sciiv.s.keyId |= MASK_EXT_IV;     /* Set ExtIV(bit 5) flag on */
    tfcb_p->iv = tfcb_p->u_sciiv.sciIv;
    tfcb_p->extendedIv = tfcb_p->sciExtendedIv;
    DBGPRINT(RSN, LOUD, (" IV = 0x%08X\n", tfcb_p->u_sciiv.sciIv));
    DBGPRINT(RSN, LOUD, (" ExtIV = 0x%08X\n", tfcb_p->sciExtendedIv));

    tfcb_p->dmaFifoHeaderLen += 2;    /* size in unit of DWORD(4 bytes) */
    DBGPRINT(RSN, LOUD, (" Fifo Length = %d\n", tfcb_p->dmaFifoHeaderLen));

    tfcb_p->secOpMode = CIPHER_SUITE_TKIP;
    DBGPRINT(RSN, LOUD, (" Cipher Suite = 0x%02X\n", tfcb_p->secOpMode));

    /* fill up the Encryption Key(16 bytes) */
    DBGPRINT(RSN, LOUD, (" Tfcb_p->encryptionKey ==>\n"));
    for (i = 0; i < 16; i++) {
        tfcb_p->encryptionKey[i] = *(key_p + i);
        DBGPRINT_RAW(RSN, LOUD, ("[%02X]", tfcb_p->encryptionKey[i]));
    }
    DBGPRINT(RSN, LOUD, ("\n"));

    /* MPDU Length + IV(4) + ExtIV(4) + ICV(4) */
    tfcb_p->mpduLen += 12;
    DBGPRINT(RSN, LOUD, (" MPDU Length = %d\n", tfcb_p->mpduLen));

    return TKIP_ENCAPSULATE_SUCCESS;
}       /* End of tkipHwMpduEncapsulate */
