/*
** $Id: domain.h 1.6 2004/09/03 17:43:58 tplee Exp $
*/

/******************************************************************************
Copyright (c) 2004 Inprocomm, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Inprocomm, Inc.
******************************************************************************/

#ifndef _DOMAIN_H
#define _DOMAIN_H

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/
/* Maximum number of supported subbands in a regulatory domain */
#define MAX_NUM_SUPPORTED_SUBBANDS      8


/* ISO/IEC 3166-1 two-character country codes */
#define COUNTRY_CODE_US     (((UINT_16) 'U' << 8) | (UINT_16) 'S')
#define COUNTRY_CODE_CA     (((UINT_16) 'C' << 8) | (UINT_16) 'A')
#define COUNTRY_CODE_EU     (((UINT_16) 'E' << 8) | (UINT_16) 'U')
#define COUNTRY_CODE_ES     (((UINT_16) 'E' << 8) | (UINT_16) 'S')
#define COUNTRY_CODE_FR     (((UINT_16) 'F' << 8) | (UINT_16) 'R')
#define COUNTRY_CODE_JP     (((UINT_16) 'J' << 8) | (UINT_16) 'P')

#define COUNTRY_CODE_AT     (((UINT_16) 'A' << 8) | (UINT_16) 'T')
#define COUNTRY_CODE_BE     (((UINT_16) 'B' << 8) | (UINT_16) 'E')
#define COUNTRY_CODE_DK     (((UINT_16) 'D' << 8) | (UINT_16) 'K')
#define COUNTRY_CODE_FI     (((UINT_16) 'F' << 8) | (UINT_16) 'I')
#define COUNTRY_CODE_DE     (((UINT_16) 'D' << 8) | (UINT_16) 'E')
#define COUNTRY_CODE_GR     (((UINT_16) 'G' << 8) | (UINT_16) 'R')
#define COUNTRY_CODE_IS     (((UINT_16) 'I' << 8) | (UINT_16) 'S')
#define COUNTRY_CODE_IE     (((UINT_16) 'I' << 8) | (UINT_16) 'E')
#define COUNTRY_CODE_IT     (((UINT_16) 'I' << 8) | (UINT_16) 'T')
#define COUNTRY_CODE_LI     (((UINT_16) 'L' << 8) | (UINT_16) 'I')
#define COUNTRY_CODE_LU     (((UINT_16) 'L' << 8) | (UINT_16) 'U')
#define COUNTRY_CODE_NL     (((UINT_16) 'N' << 8) | (UINT_16) 'L')
#define COUNTRY_CODE_NO     (((UINT_16) 'N' << 8) | (UINT_16) 'O')
#define COUNTRY_CODE_PT     (((UINT_16) 'P' << 8) | (UINT_16) 'T')
#define COUNTRY_CODE_SE     (((UINT_16) 'S' << 8) | (UINT_16) 'E')
#define COUNTRY_CODE_CH     (((UINT_16) 'C' << 8) | (UINT_16) 'H')
#define COUNTRY_CODE_GB     (((UINT_16) 'G' << 8) | (UINT_16) 'B')

#define COUNTRY_CODE_AU     (((UINT_16) 'A' << 8) | (UINT_16) 'U')
#define COUNTRY_CODE_CN     (((UINT_16) 'C' << 8) | (UINT_16) 'N')
#define COUNTRY_CODE_IL     (((UINT_16) 'I' << 8) | (UINT_16) 'L')
#define COUNTRY_CODE_JO     (((UINT_16) 'J' << 8) | (UINT_16) 'O')
#define COUNTRY_CODE_KR     (((UINT_16) 'K' << 8) | (UINT_16) 'R')
#define COUNTRY_CODE_SG     (((UINT_16) 'S' << 8) | (UINT_16) 'G')
#define COUNTRY_CODE_TW     (((UINT_16) 'T' << 8) | (UINT_16) 'W')
#define COUNTRY_CODE_TH     (((UINT_16) 'T' << 8) | (UINT_16) 'H')


/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/
/* Supported domain information entry */
typedef struct _DOMAIN_SUBBAND_INFO {
    UINT_8      firstChannelNum;
    UINT        numChannels;
    INT_8       maxTxPowerLevel;
    UINT_32     mibBandsSupportedValue;
} DOMAIN_SUBBAND_INFO, *PDOMAIN_SUBBAND_INFO;

typedef struct _DOMAIN_INFO_ENTRY {
    UINT_16                     countryCode;
    UINT_32                     mibRegDomainValue;
    DOMAIN_SUBBAND_INFO         band24;
    DOMAIN_SUBBAND_INFO         band5[4];
} DOMAIN_INFO_ENTRY, *PDOMAIN_INFO_ENTRY;

/* Supported channel information entry */
typedef struct _CHANNEL_FREQ_ENTRY {
    UINT_8      channelNum;
    UINT_32     freqInkHz;
} CHANNEL_FREQ_ENTRY, *PCHANNEL_FREQ_ENTRY;

/* Channel list entry */
typedef struct _CHANNEL_LIST_ENTRY {
    UINT_8      num;
#if SUPPORT_802_11A
    UINT_8      band;
#endif
    INT_8       regMaxTxPower;
} CHANNEL_LIST_ENTRY, *PCHANNEL_LIST_ENTRY;

/* Channel list */
typedef struct _CHANNEL_LIST {
    UINT                len;
    CHANNEL_LIST_ENTRY  channel[MAX_NUM_OPERATING_CHANNELS];
} CHANNEL_LIST, *PCHANNEL_LIST;


/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*                    D A T A   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                 F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/


#endif /* _DOMAIN_H */

