/*
** $Id: generaldef.h 1.26 2004/09/23 19:11:13 cmchang Exp $
*/

/******************************************************************************
Copyright (c) 2002-2003 Integrated Programmable Communications, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Integrated Programmable Communications, Inc.
******************************************************************************/

#ifndef _GENERAL_DEFINITIONS_H
#define _GENERAL_DEFINITIONS_H

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/
typedef struct _chain_list_t {
    struct list_head    head;           /* list of members in this chain */
    int                 num_elem;       /* number of allocated elements */
    void                *buf;           /* if chain members are created from one buffer */
    kal_dma_addr_t      dmaAddr;        /* dma address relative to above buf */
    UINT32              buf_sz;         /* the size of above buffer to free */
} chain_list_t;

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                             M A C R O S
*************************************************************************
*/
#define kal_pci_alloc(typ,pci_p,dma_p) \
        (typ*)kal_pci_alloc_sz(pci_p,sizeof(typ),dma_p)
#define kal_pci_free(pci_p,typ,m,dma) \
        kal_pci_free_sz(pci_p,sizeof(typ),m,dma)

/* In order to support memory pool in RTOS, we create
 * kal_mempool_alloc_sz when using frequent malloc and free.
 */
#if WLAN_USE_MEMPOOL
#define kal_mempool_alloc(sz)           slab_mem_alloc(sz)
#define kal_mempool_free(p)             slab_mem_free((void*)(p))
#else
#define kal_mempool_alloc(sz)           kal_malloc(sz)
#define kal_mempool_free(p)             kal_free(p)
#endif


/************************************************************************
*               Generic Chain List (QUEUE) class & methods
*************************************************************************
*/
#include <asm/system.h>
#define chain_num_elem(chain)   (chain)->num_elem

#define init_chain_list(chain) do { \
        INIT_LIST_HEAD(&(chain)->head); \
        (chain)->num_elem = 0; \
        (chain)->buf = NULL; \
        (chain)->buf_sz = 0; \
} while (0)

#define empty_chain_check(chain)        list_empty(&(chain)->head)

#define peek_chain_member(chain,type,list_dir,__ret) do { \
        if (empty_chain_check(chain)) { \
                __ret = NULL; \
        } \
        else \
                __ret = list_entry( (chain)->head.list_dir, type, sw.link ); \
} while (0)

#define get_chain_member(chain,type,list_dir,__ret) do { \
        if (empty_chain_check(chain)) { \
                __ret = NULL; \
        } \
        else { \
                __ret = list_entry( (chain)->head.list_dir, type, sw.link ); \
                list_del( (chain)->head.list_dir ); \
                (chain)->num_elem--; \
        } \
} while (0)

#define add_chain_member(chain,type,item,list_action) do {\
        list_action( &((type*)item)->sw.link, &(chain)->head ); \
        (chain)->num_elem++; \
} while (0)

#define del_chain_member(chain,type,item) do { \
        if (!empty_chain_check(chain)) { \
                list_del( &((type*)item)->sw.link ); \
                (chain)->num_elem--; \
        } \
} while (0)

#define rm_chain_member_in_loop(chain,type,lst) do { \
        if (!empty_chain_check(chain)) { \
                /* equivalent to : list_del(lst); */ \
                /* BUT be careful, the end-loop post assignment of lst=lst->next */ \
                /*    can cause endless loop situration of this loop */ \
                lst = lst->prev; \
                list_del( lst->next ); \
                (chain)->num_elem--; \
        } \
} while (0)


#define peek_chain_head(chain,type,ret) peek_chain_member(chain,type,next,ret)
#define peek_chain_tail(chain,type,ret) peek_chain_member(chain,type,prev,ret)

#define get_chain_head(chain,type,ret)  get_chain_member(chain,type,next,ret)
#define get_chain_tail(chain,type,ret)  get_chain_member(chain,type,prev,ret)

#define add_chain_head(chain,type,item) add_chain_member(chain,type,item,list_add)
#define add_chain_tail(chain,type,item) add_chain_member(chain,type,item,list_add_tail)

/*----------------------------------------------------------------------------*/
#define alignword(a,n)    ( (((unsigned int)(a) + ((n) - 1)) / (n)) * (n) )

#ifndef dim
#define dim(x)          ( sizeof(x) / sizeof(x[0]) )
#endif

#ifndef offsetof
#define offsetof(Struct,Member) ( (unsigned int)  & ( ((Struct *) 0) -> Member ) )
#endif

#ifndef max
#define max(a,b)        (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)        (((a) < (b)) ? (a) : (b))
#endif

#define movavg(v,a)     ( ((v) + (a)*7) / 8 )

#undef BIT
#define BIT(n)          ( (UINT32)1 << (n) )
#define BITS2(m,n)      ( BIT(m) | BIT (n) )
#define BITS3(m,n,o)    ( BIT(m) | BIT (n) | BIT (o) )
#define BITS4(m,n,o,p)  ( BIT(m) | BIT (n) | BIT (o) | BIT(p) )

/* bits range: for example BITS(16,23) = 0xFF0000
 *   ==>  (BIT(m)-1)   = 0x0000FFFF     ~(BIT(m)-1)   => 0xFFFF0000
 *   ==>  (BIT(n+1)-1) = 0x00FFFFFF
 */
#define BITS(m,n)       ( ~(BIT(m)-1)  &  ((BIT(n)-1)|BIT(n)) )

#ifndef MAC2STR
#define MAC2STR(a) (a)[0], (a)[1], (a)[2], (a)[3], (a)[4], (a)[5]
#define MACSTR "%02x:%02x:%02x:%02x:%02x:%02x"
#endif

#define isdigit(c) ('0' <= (c) && (c)  <= '9')
#define isupperhex(x) (((x) >= 'A') && ((x) <= 'F'))
#define isspace(c) ((c) == ' ' || (c)  == '\t')

/************************************************************************
*                 F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*           I N L I N E   F U N C T I O N   D E F I N I T I O N S
*************************************************************************
*/

#endif  /* _GENERAL_DEFINITIONS_H */

