/*
** $Id: sdl_macsorts.h 1.57 2004/10/04 17:22:30 cmchang Exp $
*/

/******************************************************************************
Copyright (c) 2002-2003 Integrated Programmable Communications, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Integrated Programmable Communications, Inc.
******************************************************************************/

#ifndef _SDL_MACSORTS_H
#define _SDL_MACSORTS_H

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/
#define cap_CfPoll_mask         (SYN_cPollable | SYN_cPollReq)

#define sMaxAId         	128     /* 2007 is largest allowable value */
                                	/* implementation limit may be lower */
/* Number of TupleCache entries and associated index range */
/* this value is an example, TupleCache size is implementation dependent */
#define SYN_tupleCacheSize	32

/* S/W key table size */
#define IPN_KEY_TABLE_SIZE      (sMaxAId+4+8)   /* STA, Group key, WDS */

/* Number of entries in aWepKeyMappings array at this station.
 * implementation dependent value, minimum=10 (see 8.3.2). */
#define sWepKeyMappingLength	10

/* Number of entries in defragmentation array at this station. */
/* The value is implementation dependent (min=3, see 9.5). */
#define SYN_defragSize		6

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/

/*******************************************************************
* Enumerated types used within the MAC state machines
*******************************************************************/
typedef enum {  SYN_head, /* place MMPDU at head of transmit queue */
                SYN_norm } /* place MMPDU at tail of transmit queue */
        Imed_t; /* priority for queuing MMPDUs, relative to MSDUs */

typedef enum {  SYN_rts_, /* RTS frame */
                SYN_cfpBss, SYN_cfendBss, /* start/end of CFP in own BSS */
                SYN_cfpOther, SYN_cfendOther, /* start/end of CFP in other BSS */
                SYN_cswitch, /* channel switch */
                SYN_misc, /* durId from other frame types */
                SYN_nosrc } /* non-reception events */
        NavSrc_t; /* source of duration in SetNav & ClearNav signals */

typedef enum {  SYN_sta_active, SYN_power_save }
        PsMode_t; /* power save mode of a station (PsResponse signal) */

typedef enum {  SYN_awake, SYN_doze }
        PsState_t; /* power save state of this station */

typedef enum {  SYN_noerr, SYN_class2, SYN_class3 }
        StateErr_t; /* requests disasoc or deauth (MmIndicate signal) */

typedef enum {  SYN_not_auth, SYN_auth_open, SYN_auth_key, SYN_asoc, SYN_dis_asoc }
        StationState_t; /* asoc/auth state of sta (SsResponse signal) */

typedef enum {  SYN_successful, SYN_partial, SYN_retryLimit_, SYN_txLifetime_,
                SYN_atimAck, SYN_atimNak }
        TxResult_t; /* transmission attempt status (PduConfirm signal) */


/*******************************************************************
* Enumerated types used in Mac and Mlme service primitives
* !!! the prefix 'SYN_' means SDL SYNONYM, i.e., constant definitions
*******************************************************************/
typedef enum { SYN_open_system, SYN_shared_key }
        AuthType_t; /* <authentication type> parm in Mlme primitives */

typedef enum { SYN_infrastructure, SYN_independent, SYN_any_bss }
        BssType_t; /* <BSS type> parameter & BSS description element */

typedef enum { SYN_contention, SYN_contentionFree,
               SYN_imed } /* not defined in p.279, but appears in p.414 */
        CfPriority_t; /* <priority> parameter of various requests */

typedef enum { SYN_success, SYN_invalid, SYN_write_only, SYN_read_only }
        MibStatus_t; /* <status> parm of Mlme/Plme Get/Set.confirm */

typedef enum { /* SYN_success, SYN_invalid,  -- already defined before */
                SYN_timeout=2, SYN_refused, SYN_tomany_req, SYN_already_bss }
        MlmeStatus_t; /* <status> parm of Mlme operation confirm */

typedef PsMode_t  PwrSave_t; /* <power save mode> parameter of MlmePowerMgt */

typedef enum { SYN_rx_success, SYN_rx_failure }
        RxStatus_t; /* <reception status> parm of MaUnitdata indication */

typedef enum { SYN_active_scan, SYN_passive_scan }
        ScanType_t; /* <scan type> parameter of MlmeScan.request */

typedef enum { SYN_reorderable, SYN_strictlyOrdered }
        ServiceClass_t; /* <service class> parameter for MaUnitdata */

typedef enum { /* SYN_successful=0, */ SYN_retryLimit=1, SYN_txLifetime, SYN_noBss,
               SYN_excessiveDataLength, SYN_nonNullSourceRouting,
               SYN_unsupportedPriority, SYN_unavailablePriority,
               SYN_unsupportedServiceClass, SYN_unavailableServiceClass,
               SYN_unavailableKeyMapping }
        TxStatus_t; /* <transmission status> parm of MaUnitdataStatus */

typedef enum { SYN_unspec_reason=1, SYN_auth_not_valid, SYN_deauth_lv_ss,
               SYN_inactivity, SYN_ap_overload, SYN_class2_err, SYN_class3_err,
               SYN_disas_lv_ss, SYN_asoc_not_auth
               , SYN_invalid_ie=13, SYN_mic_failure, SYN_pairwise_hs_timeout,
               SYN_group_update_timeout, SYN_nonmatch_ie, SYN_invalid_mc_cipher,
               SYN_invalid_uc_cipher, SYN_invalid_AKMP, SYN_unsup_rsn_ver,
               SYN_invalid_cap, SYN_1x_auth_fail, SYN_cipher_suite_rejected
               }
        ReasonCode_t;

typedef enum { SYN_sts_successful=0, SYN_unspec_fail=1,
               SYN_unsup_cap=10, SYN_reasoc_no_asoc, SYN_fail_other,
               SYN_unsupt_alg, SYN_auth_seq_fail,
               SYN_chlng_fail, SYN_auth_timeout, SYN_ap_full, SYN_unsup_rate,
               SYN_unsupt_shrSlot=25, SYN_unsupt_DSOFDM }
        StatusCode_t;

/* TypeSubtype defines the full, 6-bit frame type identifiers. */
/* These values are useful with ftype operator of Frame sort. */
/* -- Type field: b3 b2
 * -- Subtype field: b7 b6 b5 b4 */
typedef enum { SYN_asoc_req      =0x00, /* Association request                            */
               SYN_asoc_rsp      =0x10, /* Association response                           */
               SYN_reasoc_req    =0x20, /* Reassociation request                          */
               SYN_reasoc_rsp    =0x30, /* Reassociation response                         */
               SYN_probe_req     =0x40, /* Probe request                                  */
               SYN_probe_rsp     =0x50, /* Probe response                                 */
               SYN_beacon        =0x80, /* Beacon                                         */
               SYN_atim          =0x90, /* Announcement traffic indication message (ATIM) */
               SYN_disasoc       =0xa0, /* Disassociation                                 */
               SYN_auth          =0xb0, /* Authentication                                 */
               SYN_deauth        =0xc0, /* Deauthentication                               */
               SYN_ps_poll       =0xa4, /* Power Save (PS)-Poll                           */
               SYN_rts           =0xb4, /* Request To Send (RTS)                          */
               SYN_cts           =0xc4, /* Clear To Send (CTS)                            */
               SYN_ack           =0xd4, /* Acknowledgment (ACK)                           */
               SYN_cfend         =0xe4, /* Contention-Free (CF)-End                       */
               SYN_cfend_ack     =0xf4, /* CF-End + CF-Ack                                */
               SYN_data          =0x08, /* Data                                           */
               SYN_data_ack      =0x18, /* Data + CF-Ack                                  */
               SYN_data_poll     =0x28, /* Data + CF-Poll                                 */
               SYN_data_poll_ack =0x38, /* Data + CF-Ack + CF-Poll                        */
               SYN_null_frame    =0x48, /* Null function (no data)                        */
               SYN_cfack         =0x58, /* CF-Ack (no data)                               */
               SYN_cfpoll        =0x68, /* CF-Poll (no data)                              */
               SYN_cfpoll_ack    =0x78, /* CF-Ack + CF-Poll (no data)                     */
               SYN_qos_data  = 0x88, /*QoS Data frame*/
               SYN_qos_data_null = 0xc8 /*QoS Null Data frame*/
               }
        TypeSubtype_t;

typedef enum { SYN_ft_management, SYN_ft_control, SYN_ft_data, SYN_ft_reserved }
        BasicType_t; /* BasicTypes defines the 2-bit frame type groups */

typedef enum {  SYN_eSsId=0,        /* service set identifier (0:32) */
                SYN_eSupRates=1,    /* supported rates (1:8) */
                SYN_eFhParms=2,     /* FH parameter set (5) */
                SYN_eDsParms=3,     /* DS parameter set (1) */
                SYN_eCfParms=4,     /* CF parameter set (6) */
                SYN_eTim=5,         /* Traffic Information Map (4:254) */
                SYN_eIbParms=6,     /* IBSS parameter set (2) */
                SYN_eCountryInfo=7, /* 802.11d country information */
                SYN_eCtext=16,      /* challenge text (128, see 8.1.2.2) */
                SYN_eRSN=48,        /* RSN */
                SYN_eWPA=221,       /* WPA */
                SYN_eErpInfo=42,    /* ERP information */
                SYN_eExtSupRates=50}/* Extended Supported Rates */
        ElementID_t;

typedef enum {  SYN_cEss=BIT(0),        /* ESS capability */
                SYN_cIbss=BIT(1),       /* IBSS capability */
                SYN_cPollable=BIT(2),   /* CF-pollable (sta), PC present (ap) */
                SYN_cPollReq=BIT(3),    /* not CF poll req (sta), PC polls (ap) */
                SYN_cPrivacy=BIT(4),    /* WEP required */
                SYN_cShPreamble=BIT(5), /* Short Preamble implemented */
                SYN_cPBCC=BIT(6),
                SYN_cChAgility=BIT(7),  /* channel agility */
                SYN_cShortSlot=BIT(10), /* Short slot time */
                SYN_cRSN=BIT(11),       /* RSN capability */
                SYN_cDsssOFDM=BIT(13) } /* DSSS-OFDM */
        Capability_t;

/* Generic PHY parameter element for signals related to Beacons */
/* and Probe Responses that are PHY-type independent. */
typedef UINT8           *PhyParms_t;      /* Octetstring_t */

typedef unsigned        TU_t;
typedef unsigned        Usec_t, Kusec_t;

/*******************************************************************
* IBSS parameter set sort
*******************************************************************/
typedef struct IbssParams_ {
        UINT16        atimWin;
} IbssParms_t;

/*******************************************************************
* DS parameter set sort
*******************************************************************/
typedef struct DsParms_ {
        UINT16     curChannel;   /* Current Channel (1) */
} DsParms_t;

/*******************************************************************
* CF parameter set sort
*******************************************************************/
typedef struct CfParms_ {
        int             cfpCount;       /* CfpCount field (1) */
        int             cfpPeriod;      /* CfpPeriod field (1) */
        TU_t            cfpMaxDur;      /* CfpMaxDuration field (2) */
        TU_t            cfpDurRem;      /* CfpDurRemaining field (2) */
} CfParms_t;

/*******************************************************************
* Placeholders for Mlme/Plme Get/Set Parameter Values
*******************************************************************/
/* MibAtrib (placeholder in MlmeGet/Set definitions) */
typedef char* MibAtrib_t;

/* MibValue (placeholder in MlmeGet/Set definitions) */
typedef int MibValue_t;

/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/

/*******************************************************************
* Sorts for association management at AP
*******************************************************************/
/* Station Association Record -- only used at APs */
typedef struct AsocData_ {
        TypMacAddr      adAddr;           /* address of associated station */
        PwrSave_t       adPsm;            /* power save mode of the station */

        //UINT16          adCapability;
        //UINT16          adListenInterval; /* or beacon_int for aps */
        UINT32          adAsocTickTime;   /* time of association */
        RateSet_t       adSuptRates;      /* supported rates from association request */
        RateSet_t       adBasicRates;     /* basic rates from association request */
} AsocData_t;

/*******************************************************************
* BSS description sorts
*******************************************************************/
/* BssDscr is used with MlmeScan.confirm and MlmeJoin.request */
typedef struct BssDscr_ {
        /*-----------------2002/10/1 11:03------------------
         * definitions from SDL of 8802.11 standard
         * --------------------------------------------------*/
        TypMacAddr      bdBssId;
        UINT8           bdSsId[MAX_SSID_LEN+1];   /* 1 <= length <= 32 */
        BssType_t       bdType;
        TU_t            bdBcnPer;          /* beacon period in Time Units */
        int             bdDtimPer;         /* DTIM period in beacon periods */
        UINT8           bdTstamp[8];       /* 8 Octets from ProbeRsp/Beacon */
        UINT8           bdStartTs[8];      /* 8 Octets TSF when rx Tstamp */
        PhyParms_t      bdPhyParms;        /* empty if not needed by PHY */
        CfParms_t       bdCfParms;         /* empty if not CfPollable/no PCF */
        IbssParms_t     bdIbssParms;       /* empty if infrastructure BSS */
        Capability_t    bdCap;             /* capability information */
        TypRateString   bdBrateStr;        /* BSS basic rate set */
        RateSet_t       bdBRates;          /* BSS basic rate set */

        /*-----------------2002/10/1 11:04------------------
         * definitions for IPN2120 implementation
         * --------------------------------------------------*/
        // UCHAR           ssidLen;
        ULONG localTime[2];
        UCHAR dtimPeriod;
        UCHAR curChnl;
        USHORT atimWindow;
        USHORT capInfo;
        USHORT bssBasicRateSet;
        USHORT operationalRateSet;
        UINT supRatesLength;
        UCHAR supRates[8];
        LONG rssi;
        ULONG linkQuality;
} BssDscr_t;

/*******************************************************************
* Duplicate filtering support sorts
*******************************************************************/
typedef struct Tuple_ {            /* for duplicate filtering & defragmentation */
        bool            full;      /* =true if Tuple contains valid info */
        TypMacAddr      ta;        /* transmitting station address (Addr2) */
        SeqNum_t        sn;        /* Msdu/Mmpdu sequence number */
        FragNum_t       fn;        /* most recent Mpdu fragment number */
   #if 0
        TimeStamp_t     tRx;       /* reception time (endRx of fragment) */
   #endif
} Tuple_t;

typedef Tuple_t TupleCache_t[SYN_tupleCacheSize];

/*******************************************************************
* WEP support sorts
*******************************************************************/
typedef int		KeyIndex_t;  /* 0:3 */
typedef Octetstring_t	PrngKey_t;
typedef int 		KeyMappingRange_t; /* 1:sWepKeyMappingLength */

typedef struct KeyMap_ { /* structure used for entries in KeyMapArray */
        MacAddr_t      mappedAddr;
        bool           wepOn;
        PrngKey_t      wepKey;
} KeyMap_;

/*******************************************************************
* Defragmentation support sorts
*******************************************************************/
/* The PartialSdu structure is for INCOMPLETE MSDUs/MMPDUs
 * (generically SDUs) for which at least 1 fragment has been
 * received. Unfragmented SDUs are reported upward immediately,
 * and are never stored in instances of this structure.
 */
typedef struct PartialSdu_ {
        Frame_t         *pdu;
        UINT32          eol;
        UINT16          seqCtrl;
} PartialSdu_t;

/******************************************************************************
 *      __GLOBAL__ Variables defined in macsorts package
 ******************************************************************************/
typedef struct wlan_box_ {
        /*******************************************************************
        * Intra-MAC variables (names of form mXYZ)
        *******************************************************************/
        TypRateString   mBRateStr;           /* basic rate set for this sta */
        TypMacAddr      mBssId;              /* identifier of current (I)BSS */
        UINT16          mCap;                /* capability info from MlmeJoin */
        UINT8           mAssoc;              /* =true if STA associated w/BSS */
        UINT8           mCfp;                /* =true if CF period in progress */
        int             mDtimCount;          /* =0 at Tbtt of Beacon with DTIM */
        Octet_t         mSsId[MAX_SSID_LEN+1]; /* name of the current (I)BSS */
#if 0
        bool            mActingAsAp;         /* =true if STA started BSS */
        AsocId_t        mAId;                /* AID assigned to STA by AP */
        bool            mAtimW;              /* =true if ATIM window in prog */
        bool            mBkIP;               /* =true if backoff in prog */
        bool            mDisable;            /* =true if not in any BSS; then */
                                             /* TX only sends probe_req; RX only accepts beacon, probe_rsp */
        bool            mFxIP;               /* =true during frame exchange seq */
        bool            mIbss;               /* =true if STA is member of IBSS */
        int             mListenInt;          /* beacons between wake up @TBTT */
        Time_t          mNavEnd;             /* NAV end Time, <=now when idle */
        Time_t          mNextBdry;           /* next boundary Time; =0 if none */
        Time_t          mNextTbtt;           /* Time next beacon due to occur */
        bool            mPcAvail;            /* =true if point coord in BSS */
        bool            mPcDlvr;             /* =true if CF delivery only */
        bool            mPcPoll;             /* =true if CF delivery & polling */
        Usec_t          mPdly;               /* probe delay from start or join */
        PsState_t       mPss;                /* power save state of STA */
        bool            mReceiveDTIMs;       /* =true if DTIMs received */
        bool            mRxA;                /* =true if RX indicated by PHY */

        /**********************************************************************
        * Station configuration flags (static, supplementary to MIB)
        **********************************************************************/
        bool            sCanBeAp;    /* DEFAULT=false;  =true if STA can operate as AP */
        bool            sCanBePc;    /* DEFAULT=false;  =true if AP can be Point Coord */
        bool            sCfPollable; /* DEFAULT=true;   =true if responds to CF-polls */

#endif
        /**********************************************************************
        * Properties added for IPN2120 implementation
        **********************************************************************/
        RateSet_t       mORates;             /* operational rate set */
        RateSet_t       mBRates;             /* basic rate set */
        int             mMaxORate;
        int             mMaxBRate;
} wlan_box_t;


/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/
extern TypMacAddr bcstAddr; /* broadcast address */
extern TypMacAddr nullAddr; /* null address */

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/*******************************************************************
* FRAME sort (the basic definition of fields in MAC frames)
*******************************************************************/
/* Frame is a subtype of Octetstring with operators for creating
 * MAC headers, extracting each of the header fields and some
 * management frame fields, and modifying most of these fields.
 * There are operators to create and extract management frame
 * elements, but no operators for the frame body, IV, ICV, and FCS
 * fields, which are handled directly as Octetstrings. */
/* --------------------------------------------------------------- */
#define sdl_mkLlcFrame(_this,_addr1,_addr3,_addr4,__skb)  do { \
        _this = (Frame_t*)kal_mempool_alloc(sizeof(Frame_t)); \
        if (_this != NULL) { \
        _this->frameCtrl = host2ieee16(SYN_data); \
        ipn_memcpy( _this->addr1, _addr1, MAC_ADDR_LEN ); \
        ipn_memcpy( _this->addr2, mib_import(dot11MACAddress), MAC_ADDR_LEN ); \
        ipn_memcpy( _this->addr3, _addr3, MAC_ADDR_LEN ); \
        _this->mac_ctl = MACCTL_CHECK_CIPHER | MACCTL_CHECK_SLEEP; \
        if (_addr4) { \
                ipn_memcpy( _this->addr4, _addr4, MAC_ADDR_LEN ); \
                _this->mac_ctl |= MACCTL_WDS_FRAME; \
        } \
        _this->skb = __skb; /* => frame from LLC */ \
        _this->LLCdata = kal_skb_data(__skb); /* points to frame-body */ \
        _this->LLClength = kal_skb_len(__skb); \
        } /* end of _this is not NULL */ \
} while (0)

/* For SSID Information Element:
 *     type: eSsId_t
 *     element ID: SYN_eSsId
 *     local elem pointer : eSsId_p
 * the 'len' parameter here is for the case that ELEMENT length can't be decided
 * simply by the structure, say eSsId_t, since the size of eSsId_t::ssid, should
 * be determined by the actual string length of ssid, rather the array size.
 */
#define sdl_mkElem(buf,eId,len)  do { \
        eId##_p = (eId##_t *) buf; \
        eId##_p->elemId = SYN_##eId; \
        eId##_p->length = (len); \
        buf += (len) + 2; \
} while(0)

/* This is a procedural operator is part of sort Frame. This operator extracts
 * an element from a Management frame:
 *      elem:= getElem(fr,eI)
 * Copies the info field of element with element ID eI from frame fr into elem.
 * If there is no element with the specified element ID, elem is set to 'null'.
 */
#define sdl_getElem(buf,eId)  ( { \
        if (buf[0] == SYN_##eId) { \
                eId##_p = (eId##_t *) buf; \
                buf += buf[1] + 2; \
        } \
        else \
                eId##_p = (eId##_t *) 0; \
        eId##_p; \
} )


#define sdl_setFtype(f,v)     do { (f)->frameCtrl = host2ieee16((ieee2host16(((Frame_t *)(f))->frameCtrl) & ~BITS(2,7)) |  (v));        } while (0)
#define sdl_setBasetype(f,v)  do { (f)->frameCtrl = host2ieee16((ieee2host16(((Frame_t *)(f))->frameCtrl) & ~BITS(2,3)) | ((v) << 2 )); } while (0)
#define sdl_setSeq(f,v)       do { (f)->seqCtrl = host2ieee16((ieee2host16(((Frame_t *)(f))->seqCtrl) & 0x000F) | ((v) << 4 ));       } while (0)
#define sdl_setFrag(f,v)      do { (f)->seqCtrl = host2ieee16((ieee2host16(((Frame_t *)(f))->seqCtrl) & 0xFFF0) |  (v) );             } while (0)

#define sdl_ftype(f)          ((ieee2host16((f)->frameCtrl) & BITS(2,7)))
#define sdl_basetype(f)       ((ieee2host16((f)->frameCtrl) & BITS(2,3)) >> 2)
#define sdl_seq(f)             (ieee2host16((f)->seqCtrl) >> 4)
#define sdl_frag(f)            (ieee2host16((f)->seqCtrl) & 0x0F)

#define sdl_setToDs(f,b)      do { (f)->frameCtrl = host2ieee16((ieee2host16((f)->frameCtrl) & ~BIT(8 )) | ((b) << 8 )); } while (0)
#define sdl_setFrDs(f,b)      do { (f)->frameCtrl = host2ieee16((ieee2host16((f)->frameCtrl) & ~BIT(9 )) | ((b) << 9 )); } while (0)
#define sdl_setMoreFrag(f,b)  do { (f)->frameCtrl = host2ieee16((ieee2host16((f)->frameCtrl) & ~BIT(10)) | ((b) << 10)); } while (0)
#define sdl_setRetryBit(f,b)  do { (f)->frameCtrl = host2ieee16((ieee2host16((f)->frameCtrl) & ~BIT(11)) | ((b) << 11)); } while (0)
#define sdl_setPwrMgt(f,b)    do { (f)->frameCtrl = host2ieee16((ieee2host16((f)->frameCtrl) & ~BIT(12)) | ((b) << 12)); } while (0)
#define sdl_setMoreData(f,b)  do { (f)->frameCtrl = host2ieee16((ieee2host16((f)->frameCtrl) & ~BIT(13)) | ((b) << 13)); } while (0)
#define sdl_setWepBit(f,b)    do { (f)->frameCtrl = host2ieee16((ieee2host16((f)->frameCtrl) & ~BIT(14)) | ((b) << 14)); } while (0)
#define sdl_setOrderBit(f,b)  do { (f)->frameCtrl = host2ieee16((ieee2host16((f)->frameCtrl) & ~BIT(15)) | ((b) << 15)); } while (0)

#define sdl_toDs(sdu)         (ieee2host16((sdu)->frameCtrl) & BIT(8 ))
#define sdl_frDs(sdu)         (ieee2host16((sdu)->frameCtrl) & BIT(9 ))
#define sdl_isWDS(sdu)        (sdl_toDs(sdu) && sdl_frDs(sdu))
#define sdl_moreFrag(sdu)     (ieee2host16((sdu)->frameCtrl) & BIT(10))
#define sdl_retryBit(sdu)     (ieee2host16((sdu)->frameCtrl) & BIT(11))
#define sdl_pwrMgt(sdu)       (ieee2host16((sdu)->frameCtrl) & BIT(12))
#define sdl_moreData(sdu)     (ieee2host16((sdu)->frameCtrl) & BIT(13))
#define sdl_wepBit(sdu)       (ieee2host16((sdu)->frameCtrl) & BIT(14))
#define sdl_orderBit(sdu)     (ieee2host16((sdu)->frameCtrl) & BIT(15))
#define sdl_dataQos(sdu)      ((ieee2host16((sdu)->frameCtrl) & (BITS(2,3)|BIT(7))) == (BIT(3)|BIT(7)))


/* This procedural operator is part of sort DefragArray.
        array:= ArAge(ar,age)
   frees entry!eol < age.
   This is used both for the aging functionand to clear the DefragArray.
   -----------------------------------
        fpar  ar DefragArray,
              age Time ;
 */
#define sdl_release_ArAge(ar,tnow)  do { int __i; \
        for (__i=0; __i<SYN_defragSize; __i++) \
                if (ar[__i].pdu && ((ar[__i].eol - tnow) > 0x7fffffff \
                    || ar[__i].eol == tnow)){ \
                        sdlu_release_sdu(ar[__i].pdu); \
                        ar[__i].pdu = NULL; \
                } \
} while (0)

/* This procedural operator is part of the sort DefragArray.
        index:= ArFree(ar)
   returns index of an unused entry in the array.
   If all entries are used, either returns -1, or selects an
   arbitrary entry to free in order to return a usable index. Decision
   criteria for case of no free entries are implementation dependent.
 */
#define sdl_ArFree(ar,_ret)  do { int __i; \
        _ret = -1; \
        for (__i=0; __i<SYN_defragSize; __i++) \
                if (!ar[__i].pdu) \
                {   _ret = __i;  break;  } \
} while (0)

/* This procedural operator is part of sort DefragArray.
        index:= ArSearch(ar, addr, seq, frag)
   where array is a DefragArray; index is returned to identify the first element
   for which (inUse=true) and (entry!rta=addr) and (entry!rsn=seq) and (entry!rCur=frag-1);
   index is returned =1 if no match is found.
   -----------------------------------
        fpar    ar DefragArray,
                adr MacAddr,
                seq SeqNum,
                frg FragNum ;
 */
#define sdl_ArSearch(ar,addr,seq,frag,_ret) do { int __i; \
        _ret = -1; \
        for (__i=0; __i<SYN_defragSize; __i++) { \
                if (ar[__i].pdu && sdl_macEqu(ar[__i].pdu->addr2,addr) \
                    && (ar[__i].seqCtrl>>4)==seq \
                    && ((ar[__i].seqCtrl+1)&0xf) == frag) \
                {   _ret = __i;  break;  } \
        } \
} while (0)

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*           I N L I N E   F U N C T I O N   D E F I N I T I O N S
*************************************************************************
*/

#endif /* _SDL_MACSORTS_H */
