/*
** $Id: wlanmac.h 1.83 2004/10/04 17:23:07 cmchang Exp $
*/

/******************************************************************************
Copyright (c) 2002-2003 Integrated Programmable Communications, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Integrated Programmable Communications, Inc.
******************************************************************************/

#ifndef _WLANMAC_H
#define _WLANMAC_H

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/* Default values for 802.11 MIB resource attributes */
#define DEFAULT_MANUFACTURER_OUI            {0x00, 0x08, 0x22}
#define DEFAULT_MANUFACTURER_NAME           "Integrated Programmable Communications, Inc."
#define DEFAULT_MANUFACTURER_PRODUCT_NAME   "InProComm 2220 PCI 802.11g WLAN Adapter"
#define DEFAULT_MANUFACTURER_PRODUCT_VER    "Version 0.1"

/* Default SSID */
#define DEFAULT_SSID            "Venus"
#define DEFAULT_SSID_LEN        5

/* Max number of multiple SSID */
#define MAX_SSID                8

#define MAX_SSID_LEN            32
#define WLAN_NUM_RATE_SET       4
#define WLAN_MAX_RATE_STR       (MAX_NUM_SUPPORTED_DATA_RATES+1)
#define WLAN_MAX_SEQNUM         4096


/* The weighting coefficient of the MPDU retry count in the latest
 * transmission for the dynamic rate switching mechanism.
 */
#define RATE_SWITCH_LATEST_COUNT_WEIGHTING  2

/* The initial value of the weighted MPDU retry count for a new rate changed
 * by the dynamic rate switching mechanism.
 */
#define RATE_SWITCH_INITIAL_WEIGHTED_COUNT  38

/* The thresholds for the rate switching mechanism. */
#define RATE_SWITCH_UPGRADE_THRESHOLD       3
#define RATE_SWITCH_FALLBACK_THRESHOLD      34

/*---------------------------------------------------------------
* Named static data values (names of form sXYZ)
*---------------------------------------------------------------*/
#define sMaxMsduLng     2304    /* max octets in an MSDU */
#define sMacHdrLng      24      /* octets in data header, no WEP */
#define sWepHdrLng      28      /* octets in data header with WEP */
#define sWepAddLng      8       /* octets added for WEP */
#define sWdsAddLng      6       /* octets added for WDS (addr4) */
#define sCrcLng         4       /* octets for crc32 (FCS, ICV) */
#define sMaxMpduLng             /* max octets in an MPDU */ \
                        (sMaxMsduLng + sMacHdrLng + sWdsAddLng + sWepAddLng + sCrcLng)
#define sTsOctet        24      /* first octet of Timestamp field */
#define sMinFragLng     256     /* min value for aMpduMaxLength */
#define sMaxFragNum             /* maximum fragment number, = 10 */ \
                        (sMaxMsduLng / (sMinFragLng - sMacHdrLng - 6 - sCrcLng))
#define sAckCtsLng      112     /* bits in ACK and CTS frames */
#define sChallengeLng   128

#define WLAN_MAC_FCS_LEN        4    /* length of FCS field in MAC frame */
#define WLAN_MAC_IV_LEN         4    /* length of IV field in encrypted MAC frame */
#define WLAN_MAC_ICV_LEN        4    /* length of ICV field in encrypted MAC frame */
#define WLAN_MAC_EXT_IV_LEN     4    /* length of Extented ICV field in encrypted MAC frame */
#define WLAN_MAC_MIC_LEN        8    /* length of MIC field in encrypted MAC frame */
#define WLAN_MAC_REPLAY_CTR_LEN 4    /* length of Replay Counter field in encrypted MAC frame */

/* flag definitions for mac_tx_descript_t::mac_ctl */
#define MACCTL_SDU_FRAME_IN_SKB         BIT(0)   /* 1/0 frame in skb - relate to RFB */
#define MACCTL_WDS_FRAME                BIT(1)   /* 1/0 WDS frame, i.e., include ADDR4 */
#define MACCTL_NEED_TIMESTAMP           BIT(2)   /* 1/0 frame will be inserted time stamp */
#define MACCTL_QOS_FRAME                BIT(3)   /* 1/0 frame is QoS frame */
#define MACCTL_CHECK_SLEEP              BIT(4)   /* 1/0 Need to check power-saving state */
#define MACCTL_CHECK_CIPHER             BIT(5)   /* 1/0 Need to check security */

/************************************************************************
*                         D A T A   T Y P E S
*          < according TO standard 'macsort' >
*************************************************************************
*/
/*-----------------2002/10/1 13:35------------------
 * global variable of type i2220mgmt_t, which is often
 * referenced as local 'priv_p' elsewhere, but is globally
 * accessed within SDL related codes, to reduce parameter
 * passing (stack frame) overhead.
 * --------------------------------------------------*/
#define mib_import(m)           priv_p->mib.m

typedef short   FragNum_t;      /* Range of possible fragment numbers, 0:sMaxFragNum */
typedef short   SeqNum_t;       /* Range of possible sequence numbers, 0:4095 */

/*---------------------------------------------------------------
*        OCTET sort    (influenced by Z.105, Annex A)
*---------------------------------------------------------------*/
/* Octet is a subtype of Bitstring where length always =8. */
typedef UINT8           Octet_t;

/*---------------------------------------------------------------
*        OCTET STRING sort    (somewhat influenced by Z.105, Annex A)
*---------------------------------------------------------------*/
/* Octetstrings are 0-ORIGIN strings of Octet, NOT 1-ORIGIN */
/* strings like Octet_String in Z.105 (hence the name change). */
typedef UINT8*          Octetstring_t;

typedef Octetstring_t           MacAddr_t;
typedef Octet_t                 TypMacAddr[MAC_ADDR_LEN];

/*---------------------------------------------------------------
*        Multi-rate support sorts
*---------------------------------------------------------------*/
typedef Octet_t                 Rate_t;
typedef Octetstring_t           RateString_t;
typedef Octet_t                 TypRateString[WLAN_MAX_RATE_STR];
typedef UINT32                  RateSet_t;

/* This struct is the same with struct ethhdr in Linux */
__KAL_PACK_BEGIN__
typedef struct ethhdr_s {
        UINT8   h_dest[MAC_ADDR_LEN];
        UINT8   h_source[MAC_ADDR_LEN];
        UINT16  h_proto;
} __KAL_ATTRIB_PACKED__ ethhdr_t;
__KAL_PACK_END__

__KAL_PACK_BEGIN__
typedef struct vlan_ethhdr_s {
        unsigned char        h_dest[MAC_ADDR_LEN];          /* destination eth addr      */
        unsigned char        h_source[MAC_ADDR_LEN];        /* source ether addr */
        unsigned short       h_vlan_proto;              /* Should always be 0x8100 */
        unsigned short       h_vlan_TCI;                /* Encapsulates priority and VLAN ID */
        unsigned short       h_vlan_encapsulated_proto; /* packet type ID field (or len) */
}__KAL_ATTRIB_PACKED__ vlan_ethhdr_t;
__KAL_PACK_END__

/*---------------------------------------------------------------
*        Frame_t type and 802.11 frames
*---------------------------------------------------------------*/
#define wlanhead_len(sdu)  ( (sdu->mac_ctl & MACCTL_WDS_FRAME) ? 30 : 24 )
#define wlanheadQos_len(sdu)  ( (sdu->mac_ctl & MACCTL_WDS_FRAME) ? 32 : 24 )
/*-----------------2002/10/1 16:22------------------
 * IEEE 802.11 WLAN Frame Structure, anonymous structure
 * This is to provide a C++ base / derived class framework
 * reserved[2] for IP layer alignment
 * --------------------------------------------------*/
#define __mac_header_standard_t \
        UINT16 frameCtrl; \
        UINT16 durationID; \
        UINT8  addr1[MAC_ADDR_LEN]; \
        UINT8  addr2[MAC_ADDR_LEN]; \
        UINT8  addr3[MAC_ADDR_LEN]; \
        UINT16 seqCtrl;  /* Sequence Control */ \
        UINT8  addr4[MAC_ADDR_LEN]; \
        UINT16 qosCtrl

/* MAC control descriptor type
 * NOTE that the LLCdata pointer for TX within SDL, excludes the wlan mac header
 * while for RX, includes the mac header
 * ---------------------------------------------------------------
 * How will skb be treated ? For data frame from LLC, the ownership of
 * skb is passed down from upper to driver and is released after TDMA done.
 * For data frame from STA-A to STA-B, the skb is allocated by driver
 * and will also released by driver.
 * However, for management frame generated within driver, the skb is
 * _NULL_, but the memory pointed by _LLCdata_ must be released when TDMA done.
 */
#define __mac_ctl_descript_t \
        UINT16 mac_ctl;  /* control flag */ \
        UINT16 LLClength;/* LLCdata length in LLC MA-UNITDATA.request message */ \
        UINT8  *LLCdata; /* LLCdata pointer */ \
        kal_skb_t *skb

#define mac_header_common_t \
        __mac_ctl_descript_t; \
        __mac_header_standard_t

#define wlanheader_offset       offsetof(Frame_t,frameCtrl)
#define mac_ctl_descript_sz     wlanheader_offset
#define wlanbody_offset         (offsetof(Frame_t,qosCtrl)+sizeof(UINT16))

/* Type cast to wlan_mac_header_t type */
#define tcast_wlan_header(fr) \
        ((wlan_mac_header_t*)( &((Frame_t*)(fr))->frameCtrl ))

/* IEEE 802.11 WLAN Frame Structure */
/* WLAN 24-byte MAC Header (without Address 4 field) */
__KAL_PACK_BEGIN__
typedef struct wlan_mac_header_  {
        __mac_header_standard_t;
} __KAL_ATTRIB_PACKED__ wlan_mac_header_t;
__KAL_PACK_END__

/* IPN2220 SDL Hypothetical Architecture internal Frame Structure *
 * Every MAC frame intra-communicated within the driver has this
 * structure as the base type. You can say every other frame is derived
 * (or inherited) from this base class, if in C++ terms.
 */
__KAL_PACK_BEGIN__
typedef struct Frame_t_  {
        mac_header_common_t;
} __KAL_ATTRIB_PACKED__ Frame_t;
__KAL_PACK_END__

/* When the mac_header length is 0 (exception case, control frame),
 * 32 bytes are needed for bubble. In general, the bubble size is 8 bytes.
 * So here we only reserve 8 bubble size in the above rxBufferData[]
 * when we only receive data and management frame.
 */
#define MAX_RX_PKT_LEN          (RX_SKB_BUF_SZ - wlanheader_offset - 8)

/* WLAN buffer (including MAC header) for receive, maximum possible size */
__KAL_PACK_BEGIN__
typedef struct rxbuf_frame_  {
        mac_header_common_t;
        UINT8       rxBufferData[RX_SKB_BUF_SZ - wlanbody_offset];
} __KAL_ATTRIB_PACKED__ rxbuf_frame_t;
__KAL_PACK_END__

/* Beacon/Probe Response frame format */
__KAL_PACK_BEGIN__
typedef struct _BEACON_FRAME {
        /* Beacon MAC header */
        mac_header_common_t;

        /* Beacon/Probe Response frame body */
        ULONG   timestamp[2];           /* Timestamp */
        UINT16  beaconInterval;         /* Beacon interval */
        UINT16  capInfo;                /* Capability information */
        UCHAR   infoElem[168];          /* Information elements, include RSN IE */
                                        /* 168 (172-4) is for max size in chip */
} __KAL_ATTRIB_PACKED__ beacon_frame_t;
__KAL_PACK_END__

typedef beacon_frame_t probe_rsp_frame_t;

/* Probe Request frame format */
__KAL_PACK_BEGIN__
typedef struct _PROBE_REQ_FRAME {
        /* Probe Request MAC header */
        mac_header_common_t;

        /* Probe Request frame body */
        UINT8   infoElem[58];       /* Information elements */
} __KAL_ATTRIB_PACKED__ probe_req_frame_t;
__KAL_PACK_END__

/* IBSS ATIM frame format */
__KAL_PACK_BEGIN__
typedef struct _ATIM_FRAME {
        /* IBSS ATIM MAC header */
        mac_header_common_t;

        /* IBSS ATIM frame body -- null */
} __KAL_ATTRIB_PACKED__ atim_frame_t;
__KAL_PACK_END__

/* Disassociation frame format */
__KAL_PACK_BEGIN__
typedef struct _DISASSOC_FRAME {
        /* Disassociation MAC header */
        mac_header_common_t;

        /* Disassociation frame body */
        UINT16  reasonCode;             /* Reason code */
} __KAL_ATTRIB_PACKED__ disassoc_frame_t;
__KAL_PACK_END__

/* Association Request frame format */
__KAL_PACK_BEGIN__
typedef struct _ASSOC_REQ_FRAME {
        /* Association Request MAC header */
        mac_header_common_t;

        /* Association Request frame body */
        UINT16  capInfo;                /* Capability information */
        UINT16  listenInterval;         /* Listen interval */
        UINT8   infoElem[58];           /* Information elements */
} __KAL_ATTRIB_PACKED__ assoc_req_frame_t;
__KAL_PACK_END__

/* Association Response frame format */
__KAL_PACK_BEGIN__
typedef struct _ASSOC_RSP_FRAME {
        /* Association Response MAC header */
        mac_header_common_t;

        /* Association Response frame body */
        UINT16  capInfo;                /* Capability information */
        UINT16  statusCode;             /* Status code */
        UINT16  assocId;                /* Association ID */
        UINT8   supRates[24];           /* Supported rates */
} __KAL_ATTRIB_PACKED__ assoc_rsp_frame_t;
__KAL_PACK_END__

/* Reassociation Request frame format */
__KAL_PACK_BEGIN__
typedef struct _REASSOC_REQ_FRAME {
        /* Reassociation Request MAC header */
        mac_header_common_t;

        /* Reassociation Request frame body */
        UINT16  capInfo;                /* Capability information */
        UINT16  listenInterval;         /* Listen interval */
        UINT8   currentApAddr[MAC_ADDR_LEN];    /* Current AP address */
        UINT8   infoElem[60];           /* information elements */
} __KAL_ATTRIB_PACKED__ reassoc_req_frame_t;
__KAL_PACK_END__

/* Reassociation Response frame format
   (the same as Association Response frame) */
typedef assoc_rsp_frame_t reassoc_rsp_frame_t;

/* Authentication frame format */
__KAL_PACK_BEGIN__
typedef char    challenge_t[sChallengeLng];
typedef struct _AUTH_FRAME {
        /* Authentication MAC header */
        mac_header_common_t;

        /* Authentication frame body */
        UINT16  authAlgNum;             /* Authentication algorithm number */
        UINT16  authTransSeqNo;         /* Authentication transaction sequence no. */
        UINT16  statusCode;             /* Status code */
        UINT8   ElementID;              /* Information Elements */
        UINT8   Length;                 /* the challengeText length */
        UINT8   challengeText[sChallengeLng];   /* Challenge text */
} __KAL_ATTRIB_PACKED__ auth_frame_t;
__KAL_PACK_END__

/* Deauthentication frame format */
__KAL_PACK_BEGIN__
typedef struct _DEAUTH_FRAME {
        /* Deauthentication MAC header */
        mac_header_common_t;

        /* Deauthentication frame body */
        UINT16  reasonCode;             /* Reason code */
} __KAL_ATTRIB_PACKED__ deauth_frame_t;
__KAL_PACK_END__


/* ----- Information Elements ----- */
/* SSID element format */
__KAL_PACK_BEGIN__
typedef struct _SSID_ELEM {
        UINT8   elemId;                 /* Element ID */
        UINT8   length;                 /* Length */
        UINT8   ssid[1];                /* SSID, RANGE 0~32 */
} __KAL_ATTRIB_PACKED__ eSsId_t;
__KAL_PACK_END__

/* Supported Rates element format */
__KAL_PACK_BEGIN__
typedef struct _SUPPORTED_RATES_ELEM {
        UINT8   elemId;                 /* Element ID */
        UINT8   length;                 /* Length */
        UINT8   supRates[1];            /* Supported Rates, RANGE 1~8 */
} __KAL_ATTRIB_PACKED__ eSupRates_t;
__KAL_PACK_END__

/* DS Parameter Set element format */
__KAL_PACK_BEGIN__
typedef struct _DS_PARAM_SET_ELEM {
        UINT8   elemId;                 /* Element ID */
        UINT8   length;                 /* Length */
        UINT8   curChnl;                /* Current Channel */
} __KAL_ATTRIB_PACKED__ eDsParms_t;
__KAL_PACK_END__

/* CF Parameter Set element format */
__KAL_PACK_BEGIN__
typedef struct _CF_PARAM_SET_ELEM {
        UINT8   elemId;                 /* Element ID */
        UINT8   length;                 /* Length */
        UINT8   cfpCount;               /* CFP Count */
        UINT8   cfpPeriod;              /* CFP Period */
        UINT16  cfpMaxDuration;         /* CFP MaxDuration */
        UINT16  cfpDurRemaining;        /* CFP DurRemaining */
} __KAL_ATTRIB_PACKED__ eCfParms_t;
__KAL_PACK_END__

/* TIM element format */
__KAL_PACK_BEGIN__
typedef struct _TIM_ELEM {
        UINT8   elemId;                 /* Element ID */
        UINT8   length;                 /* Length */
        UINT8   dtimCount;              /* DTIM Count */
        UINT8   dtimPeriod;             /* DTIM Period */
        UINT8   bitmapCtrl;             /* Bitmap Control */
        UINT8   bitmap[251];            /* Partial Virtual Bitmap, RANGE 1~251 */
} __KAL_ATTRIB_PACKED__ eTim_t;
__KAL_PACK_END__

/* IBSS Parameter Set element format */
__KAL_PACK_BEGIN__
typedef struct _IBSS_PARAM_SET_ELEM {
        UINT8   elemId;                 /* Element ID */
        UINT8   length;                 /* Length */
        UINT8   atimWindow[2];          /* ATIM Window */
} __KAL_ATTRIB_PACKED__ eIbParms_t;
__KAL_PACK_END__

/* Challenge Text element format */
__KAL_PACK_BEGIN__
typedef struct _CHALLENGE_TEXT_ELEM {
        UINT8   elemId;                 /* Element ID */
        UINT8   length;                 /* Length */
        UINT8   challengeText[1];       /* Challenge Text, RANGE 1~253, normal 128 */
} __KAL_ATTRIB_PACKED__ eCtext_t;
__KAL_PACK_END__

/* ERP Information element format */
__KAL_PACK_BEGIN__
typedef struct _ERP_INFO_ELEM {
        UINT8   elemId;                 /* Element ID */
        UINT8   length;                 /* Length */
        UINT8   info;                   /* ERP information (bitwise) */
} __KAL_ATTRIB_PACKED__ eErpInfo_t;
__KAL_PACK_END__

/* Extended Supported Rates element format */
__KAL_PACK_BEGIN__
typedef struct _EXT_SUPPORTED_RATES_ELEM {
        UINT8   elemId;                 /* Element ID */
        UINT8   length;                 /* Length */
        UINT8   supRates[1];            /* Supported Rates, RANGE 1~255 */
} __KAL_ATTRIB_PACKED__ eExtSupRates_t;
__KAL_PACK_END__

/* WPA Information element format for association/reassociation requesst */
__KAL_PACK_BEGIN__
typedef struct _WPA_INFO_ELEM {
        UINT8   elemId;
        UINT8   length;
        UINT8   oui[4];
        UINT16  version;
        UINT8   groupKeyCipherSuite[4];
        UINT16  pairwiseKeyCipherSuiteCount;
        struct {
                UINT8   pairwiseKeyCipherSuite[4];
        } pairwiseKeyCipherSuiteList[3]; /* WEP, TKIP and CCMP */
        UINT16 authKeyMgtSuiteCount;
        struct {
                UINT8  authKeyMgtSuite[4];
        } authKeyMgtSuiteList[1];
        UINT16 capability;
} __KAL_ATTRIB_PACKED__ eWPA_t;
__KAL_PACK_END__

/* WPA Information element format for association/reassociation requesst */
__KAL_PACK_BEGIN__
typedef struct _REQ_WPA_INFO_ELEM {
        UCHAR   elemId;
        UCHAR   length;
        UCHAR   oui[4];
        USHORT  version;
        UCHAR   groupKeyCipherSuite[4];
        USHORT  pairwiseKeyCipherSuiteCount;
        UCHAR   pairwiseKeyCipherSuite[4];
        USHORT  authKeyMgtSuiteCount;
        UCHAR   authKeyMgtSuite[4];
        USHORT  cap;
} __KAL_ATTRIB_PACKED__ eREQ_WPA_t;
__KAL_PACK_END__

/* Country Information element format */
__KAL_PACK_BEGIN__
typedef struct _COUNTRY_INFO_ELEM {
        UINT8   elemId;                 /* Element ID */
        UINT8   length;                 /* Length */
        UINT8   countryString[3];       /* Country Code */
        struct {
            UINT8       firstChannelNum;
            UINT8       numChannels;
            UINT8       maxTxPowerLevel;
        } subBands[5]; /* Sub-bands information */
} __KAL_ATTRIB_PACKED__ eCountryInfo_t;
__KAL_PACK_END__

/*******************************************************************
* Fragmentation support sorts
*******************************************************************/
/* Array to hold up to FragNum fragments of an Msdu/Mmpdu */
//typedef Frame_t       FragMpdu_t;
typedef struct FragMpdu_s FragMpdu_t;
struct FragMpdu_s {
        UINT8           *LLCdata;
        UINT32          LLClength;      /* for alignment, 32bits instead of 16bits */
};
typedef FragMpdu_t  FragArray_t[sMaxFragNum];

/* FragSdu structure is for OUTGOING MSDUs/MMPDUs (called SDUs) */
/* Each SDU, even if not fragmented, is held in an instance of */
/* this structure awaiting its (re)transmission attempt(s). */
/* Transmit queue(s) are ordered lists of FragSdu instances. */
typedef struct FragSdu_ FragSdu_t;
struct FragSdu_ {
        /*------------------------------------------
         * properties added as implementation required
         *------------------------------------------*/
        struct {
            struct list_head    link;   /* linkage for SduQueue_t */
            Frame_t             *sdu;
            sta_info_t          *sta_p;
        #if SUPPORT_PERF_MEASURE
            UINT32              txProcTime1, txProcTime2, txProcTime3;
        #endif
        } sw;
        UINT8           encryptKey[32];
        UINT8           cipherSuite;
        UINT8           encryptionKeyId;

        UINT8           grpa;     /* =true if RA (not DA) is a group address */
        UINT8           cf;       /* requested priority (from LLC, TYPE CfPriority_t */

        /*------------------------------------------
         * properties of standard defined
         *------------------------------------------*/
        FragNum_t       fTot;     /* number of fragments in pdus FragArray */
        UINT32          eol;      /* set to (now + dUsec(aMaxTxMsduLifetime))
                                    when the entry is created */
#if SUPPORT_TKIP
        int             mod4mic;  /* detail see the dot11.c sdlu_fragment func */
        UINT8           tkipMic[8];
        UINT8           *da_p;
        UINT8           *sa_p;
        UINT16          lastfTot;
#endif
        //FragMpdu_t      pdus[0];  /* array to hold fragments of a frame */
};

/* Get pointer to appended FragMpdu_t array */
#define fsdu_pdu(fsdu,n)        (((FragMpdu_t*)((FragSdu_t*)(fsdu) + 1)) + (n))

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/* Endian swap between 802.11 field and host variable */
#define ieee2host16(n)          chip2host_16(n)
#define ieee2host32(n)          chip2host_32(n)
#define host2ieee16(n)          host2chip_16(n)
#define host2ieee32(n)          host2chip_32(n)
#if 0
/*---------------------------------------------------------------
*        Time Stamp sorts
*---------------------------------------------------------------*/
#define ts_assign(a,b)           ((TimeStamp_t)(a)) = ((TimeStamp_t)(b))
#define ts_add(a,b)             (((TimeStamp_t)(a)) + ((TimeStamp_t)(b)))
#define ts_sub(a,b)             (((TimeStamp_t)(a)) - ((TimeStamp_t)(b)))
#define ts_cmp(a,b)              ts_sub(a,b)
#define ts_lowPart(t)           ((UINT32)(t))
#define ts_hiPart(t)            ((UINT32) (((TimeStamp_t)(t))>>32))
#define ts_quad(a,b)            ((((TimeStamp_t)(a))<<32) + (b))

#define ts_timeval(a)           (((TimeStamp_t)(a)->tv_sec) * 1000000 + (a)->tv_usec)
#endif
/*---------------------------------------------------------------
*        MAC Address sorts
*---------------------------------------------------------------*/
/* MacAddr is a subtype of Octetstring with added operators: */
/*    isGroup(m) =true if given a group address */
/*    isBcst(m)  =true if given the broadcast address */
/*    isLocal(m) =true if given a locally-administered address */
/*    adrOs(m)   converts MacAddr to Octetstring */
/* MAC addresses must be defined to be exactly 6 octets long, */
/* typically using the S6 operator or nullAddr synonym. */
/* --------------------------------------------------------------- */
#define sdl_isGroup(m)                  (m[0] & 0x01)
#define sdl_isBcst(m)                   (m[0] == 0xff  &&  m[1] == 0xff  &&  m[2] == 0xff &&\
                                         m[3] == 0xff  &&  m[4] == 0xff  &&  m[5] == 0xff)
#define sdl_isLocal(m)                  (m[0] & 0x02)

#define sdl_macEqu(a1,a2)       (ipn_memcmp(a1,a2,MAC_ADDR_LEN) == 0)

/*---------------------------------------------------------------
*        Multi-rate support sorts
*---------------------------------------------------------------*/
/* converts (rate,bitCount) to integer usec */
#define sdl_calcDur(r,i)        (((((10000000 + ((r & 0x7F) - 1)) / \
                                (500 * (r & 0x7F))) * i) + 9999) / 10000)
#define sdl_rateVal(r)          ((r) & 0x7F)        /* clears high-order bit */
#define sdl_isBasic(r)          ((r) & 0x80)        /* true if high-order bit set */
#define sdl_basicRate(r)        ((r) |= 0x80)       /* sets high-order bit */
#define sdl_notBasicRate(r)     ((r) &= 0x7f)       /* clear high-order bit */

/* Macros to get and set the wireless LAN frame fields those are 16/32 bits in
   length. */
__KAL_INLINE__ void WLAN_GET_FIELD_16(unsigned char* _memAddr_p,
        unsigned short* _value_p){
        unsigned short result = 0;
        result = ((UINT_16) _memAddr_p[0]) | (((UINT_16) _memAddr_p[1]) << 8);
        *_value_p = result;
}

__KAL_INLINE__ void WLAN_GET_FIELD_32(unsigned char* _memAddr_p,
        unsigned long* _value_p){
        unsigned long result = 0;
        result = ((UINT_32) _memAddr_p[0]) | (((UINT_32) _memAddr_p[1]) << 8) |
        (((UINT_32) _memAddr_p[2]) << 16) | (((UINT_32) _memAddr_p[3]) << 24);
        *_value_p = result;
}

__KAL_INLINE__ void WLAN_SET_FIELD_16(unsigned char* _memAddr_p,
        unsigned short _value){
        _memAddr_p[0] = _value &0x00FF;
        _memAddr_p[1] = (_value &0xFF00) >> 8;
}

__KAL_INLINE__ void WLAN_SET_FIELD_32(unsigned char* _memAddr_p,
        unsigned long _value){
        _memAddr_p[0] = _value &0x000000FF;
        _memAddr_p[1] = (_value &0x0000FF00) >> 8;
        _memAddr_p[2] = (_value &0x00FF0000) >> 16;
        _memAddr_p[3] = (_value &0xFF000000) >> 24;
}

#ifdef IPN_NEW_KAL
#if __CPU_ENDIAN__ == __BIG_ENDIAN__
#define kal_ntohs(v) (v)
#define kal_htons(v) (v)
#else
#define kal_ntohs(v) __swap16(v)
#define kal_htons(v) __swap16(v)
#endif
#endif
/************************************************************************
*                 F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*           I N L I N E   F U N C T I O N   D E F I N I T I O N S
*************************************************************************
*/
#endif /* _WLANMAC_H */

