/*
** $Id: 11i-rsn.h 1.10 2004/08/24 10:33:59 whsu Exp $
*/

/*!     \file 11i-rsn.h
        \brief IEEE 802.11i rsn standard SDL, Authenticator key management state machine
*/

/******************************************************************************
Copyright (c) 2000-2002 Integrated Programmable Communications, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Integrated Programmable Communications, Inc.
******************************************************************************/

/********************************************************************************************/
/*            802.1x variables                                                              */
/********************************************************************************************/
typedef enum {
    ForceUnauthorized,
    ForceAuthorized,
    Auto
} PortTypes;

/*----------------------------------------------------------------------------*/
/* if definde                                                                 */
/*----------------------------------------------------------------------------*/
#define get_wpa_oui_type(___v) (((___v) & 0xFF000000) >> 24)
#define set_wpa_oui_type(___v, ___v2) (((___v) = (((___v)& 0x00FFFFFF) | ((unsigned long)___v2) << 24)))

#define RSN_ENABLE      1
#define TKIP_ENABLE     1
#define CCMP_ENABLE     1

#define PAIRWISE_REKEY      1
#define GROUP_REKEY_TIME    2
#define EAPOL_MIC_FAILURE   3
#define DATA_MIC_FAILURE    4

#define CCM_IB_FLAGS    0x59
#define CCM_CTR_FLAGS   0x01

/* ----- Constants defined for WPA OUI ----- */
#define WPA_OUI                       0x01f25000
#define WPA_OUI_CIPHER_NONE           0x00f25000
#define WPA_OUI_CIPHER_WEP40          0x01f25000
#define WPA_OUI_CIPHER_TKIP           0x02f25000
#define WPA_OUI_CIPHER_WRAP           0x03f25000
#define WPA_OUI_CIPHER_CCMP           0x04f25000
#define WPA_OUI_CIPHER_WEP104         0x05f25000
#define WPA_OUI_CIPHER_WEP128         0x06f25000

#define TKIP_ENCAPSULATE_SUCCESS        ENCAPSULATE_SUCCESS
#define TKIP_DECAPSULATE_SUCCESS        DECAPSULATE_SUCCESS
#define TKIP_ENCRYPT_SUCCESS            ENCRYPT_SUCCESS
#define TKIP_DECRYPT_SUCCESS            DECRYPT_SUCCESS
#define TKIP_ICV_ERROR                  21
#define TKIP_MIC_ERROR                  22
#define TKIP_REPLAY_ERROR               23
#define TKIP_FRAME_FORMAT_ERROR         24
#define TKIP_NULL_KEY_ERROR             25

#define WRAP_ENCAPSULATE_SUCCESS        ENCAPSULATE_SUCCESS
#define WRAP_DECAPSULATE_SUCCESS        DECAPSULATE_SUCCESS
#define WRAP_ENCRYPT_SUCCESS            ENCRYPT_SUCCESS
#define WRAP_DECRYPT_SUCCESS            DECRYPT_SUCCESS
#define WRAP_MIC_ERROR                  32
#define WRAP_REPLAY_ERROR               33
#define WRAP_FRAME_FORMAT_ERROR         34
#define WRAP_NULL_KEY_ERROR             35

#define CCMP_ENCAPSULATE_SUCCESS        ENCAPSULATE_SUCCESS
#define CCMP_DECAPSULATE_SUCCESS        DECAPSULATE_SUCCESS
#define CCMP_ENCRYPT_SUCCESS            ENCRYPT_SUCCESS
#define CCMP_DECRYPT_SUCCESS            DECRYPT_SUCCESS
#define CCMP_MIC_ERROR                  42
#define CCMP_REPLAY_ERROR               43
#define CCMP_FRAME_FORMAT_ERROR         44
#define CCMP_NULL_KEY_ERROR             45

/***********************************************************************************************/
/*          Md5.c define header                                                                */
/***********************************************************************************************/
/* MD5 context structure */
typedef struct _MD5_CTX {
    ULONG A, B, C, D;
    ULONG X[16];
    ULONG AA, BB, CC, DD;
    ULONG bitCount[2];     /* processed bit count */
    ULONG unProcByteCount; /* number of un-processed byte count in X[] */
} MD5_CTX, *PMD5_CTX;


/***********************************************************************************************/
/*          sha1.c define header                                                               */
/***********************************************************************************************/
/* SHA-1 context structure */
typedef struct _SHA1_CTX{
    ULONG bitCount[2];     /* processed bit count */
    ULONG unProcByteCount; /* number of un-processed byte count in X[] */
    ULONG H[5];
    ULONG W[80];
    ULONG K[80];
    ULONG X[16];
    ULONG TEMP;
    ULONG A, B, C, D, E;
} SHA1_CTX, *PSHA1_CTX;


/*********************************************************************************************/
/*            11i-rsn Authenticator state machine variables                                  */
/*********************************************************************************************/
/*            Signal                                                                         */
/*********************************************************************************************/

#define SIG_rsn_Disconnect      0
#define SIG_rsn_init            1
#define SIG_AuthenticationRequest   2
#define SIG_ReAuthenticationRequest 3
#define SIG_DeauthenticationRequest 4

/* EAPoL-Key descriptor Key Information MASK */
#define BYTE_MASK_EAPOL_KI_DESCR_VER   0x07
#define BYTE_MASK_EAPOL_KI_TYPE        0x08
#define BYTE_MASK_EAPOL_KI_INDEX       0x30
#define BYTE_MASK_EAPOL_KI_INSTALL     0x40
#define BYTE_MASK_EAPOL_KI_ACK         0x80
#define BYTE_MASK_EAPOL_KI_MIC         0x01
#define BYTE_MASK_EAPOL_KI_SECURE      0x02
#define BYTE_MASK_EAPOL_KI_ERROR       0x04
#define BYTE_MASK_EAPOL_KI_REQUEST     0x08
#define BYTE_MASK_EAPOL_KI_RESERVE     0xf0

/*********************************************************************************************/
/*            11i-rsn Authenticator state machine variables                                  */
/*********************************************************************************************/
/*            State Machine & Variable                                                                       */
/*********************************************************************************************/

/* Authenticator PAE state machine */
struct Author_PTK_sm {
        /* variables */

        /* constants */

        /* counters */

    enum { // 4 Way handshack
           PTK_sm_DISCONNECTED, PTK_sm_DISCONNECT, PTK_sm_INITIALIZE, PTK_sm_AUTHENTICATION,
           PTK_sm_AUTHENTICATION2, PTK_sm_INITPMK, PTK_sm_INITPSK,
           PTK_sm_PTKSTART, PTK_sm_PTKINITNEGOTIATING, PTK_sm_PTKINITDONE,
           PTK_sm_UPDATEKEYS, PTK_sm_INTEGRITYFAILURE,PTK_sm_KEYUPDATE,
           // GK handshack
           GTK_sm_REKEYNEGOTIATING, GTK_sm_KEYERROR, GTK_sm_REKEYESTABLISHED,
           // GK update
           GTKUP_sm_SETKEYS, GTKUP_sm_SETKEYSDONE
         } state;
};

struct Author_state_machine {
        /* global variables */
        UINT8   Init;

        UINT8   Timeout4UP;
        UINT8   PRSC[6];
        UINT8   GRSC[6];

        UINT8   Disconnect;

        UINT8   EAPoLKeyReceived;
        UINT8   TimeoutEvt;
        ULONG   TimeoutCtr;
        UINT8   EAPoLOK;

        UINT8   MICVerified;
        UINT8   GTKAuthenticator;
        UINT8   IntegrityFailed;

        char    PTK[64];
        char    *MIC;
        char    *EK;
        char    *TK;

        UINT8   ANonce[32];
        //UINT8   SNonce[32];

        UINT8   GKeyDoneStations;
        UINT8   GTKRekey;
        UINT8   GInitAKeys;
        UINT8   GInitDone;
        UINT8   GUpdateStationKeys;
        UINT8   GKeyReady;

        UINT8   authSuccess;
        UINT8   portSecure;

        /* Port Timers state machine */
        /* 'Boolean tick' implicitly handled as registered timeout */
        struct Author_PTK_sm        Author_hs;//PTK_4way;

        /* Somewhat nasty pointers to global hostapd and STA data to avoid
         * passing these to every function */
        sta_info_t *sta_p;
        i2220mgmt_t *priv_p;
};


#define eophdr      0//[8]
#define eopver      8
#define eoptype     9
#define eoplen1     10
#define eoplen2     11

#define eop_len     12

#define keytype     0
#define keyInfo1    1
#define keyInfo2    2           // Key Information
#define keyLen1     3           // Key Length
#define keyLen2     4           // Key Length
#define replayCtr   5 //[8];      // Replay Counter
#define keyNonce    13//[32];     // Key Nonce
#define keyIv       45//[16];     // EAPoL-Key IV
#define keyRSC      61//[8];      // Key RSC
#define keyIde      69//[8];      // Key ID
#define keyMic      77//[16];     // Key MIC
#define keyMaterialLen1 93      // Key Material Length
#define keyMaterialLen2 94      // Key Material Length
#define keyData     95//[38];    // Key Data

__KAL_PACK_BEGIN__
typedef struct _EAPoL_KEY_MSG {
        UINT8  data[12+95+60];
/*
        UINT8  type;            // Descriptor Type
        UINT8  keyInfo1;        // Key Information
        UINT8  keyInfo2;        // Key Information
        UINT8  keyLen1;         // Key Length
        UINT8  keyLen2;         // Key Length
        UINT8  replayCtr[8];    // Replay Counter
        UINT8  keyNonce[32];    // Key Nonce
        UINT8  keyIv[16];       // EAPoL-Key IV
        UINT8  keyRSC[8];       // Key RSC
        UINT8  keyId[8];        // Key ID
        UINT8  keyMic[16];      // Key MIC
        UINT8  keyMaterialLen1; // Key Material Length
        UINT8  keyMaterialLen2; // Key Material Length
        UINT8  keyData[38];     // Key Data
*/
}__KAL_ATTRIB_PACKED__ EAPoL_KEY_MSG_t;
__KAL_PACK_END__
__KAL_PACK_BEGIN__
typedef struct Frame1xKey_t_  {
       // ieee802_1x_hdr_t   hdr;
        EAPoL_KEY_MSG_t    key;
}__KAL_ATTRIB_PACKED__ Frame1xKey_t;
__KAL_PACK_END__

typedef struct preauth_t_ {
        unsigned char macAddr[6];
        short         aid;
        unsigned char sPMKID[16];
        unsigned char iPMKID;
} preauth_t;



/*********************************************************************************************/
/*            11i-rsn Authenticator state machine variables                                  */
