/*
** $Id: wpa_kal.c,v 1.1.1.1 2005/09/15 13:37:48 root Exp $
*/

/*! \file wpa_kal.c
    \brief To provide WPA function as part of Inprocomm's
           Linux turnkey solution.
*/

/******************************************************************************
Copyright (c) 2002-2003 Inprocomm, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with Inprocomm, Inc.
******************************************************************************/

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
/*================================================================*/
/* System Includes */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/init.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/rtnetlink.h>
#include <linux/list.h>
#include <linux/if_arp.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/vmalloc.h>
#include <asm/byteorder.h>
#include <asm/uaccess.h>

#include "precomp.h"
// Sam add for wlan mac filter WLAN_MACFILTER
#include "asm/tc3162/tc3162_ipnwlan_ioc.h"
//#include <asm/tc3162/tc3162_ipnwlan_ioc.h>
// End Sam add

/* KenWang added 2005/06/21 -- begin -- */
#include <asm/siginfo.h>
#include <linux/fcntl.h>
extern fac_pid;
/* KenWang added 2005/06/21 -- end -- */

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/
static void indicate_to_1xsm(struct net_device* ap_dev, struct sk_buff* skb);
#if SUPPORT_IPNRSN
void set_wpa_uni_cipher_suite(IN PMP_ADAPTER adapter_p, IN BOOLEAN none,
                         IN BOOLEAN wep40, IN BOOLEAN tkip, IN BOOLEAN wrap,
                         IN BOOLEAN ccmp, IN BOOLEAN wep104, IN BOOLEAN wep128);
void set_rsn_uni_cipher_suite(IN PMP_ADAPTER adapter_p, IN BOOLEAN none,
                         IN BOOLEAN wep40, IN BOOLEAN tkip, IN BOOLEAN wrap,
                         IN BOOLEAN ccmp, IN BOOLEAN wep104, IN BOOLEAN wep128);
void set_sm_timer(i2220mgmt_t *priv_p);
/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
*
* \return
*/
/*----------------------------------------------------------------*/
void sm_timer_func(unsigned long data)
{
        i2220mgmt_t             *priv_p = (i2220mgmt_t*)data;
        struct list_head        *lst;
        sta_info_t              *sta_p;

        if (priv_p->mib.dot11RSNEnabled) {
                /* Note: in the following loop, we can not delete sta record */
                list_for_each(lst, &priv_p->ap.sta_chain.head) {
                        sta_p = list_entry(lst, sta_info_t, sw.link);
                        if (sta_p->aid > 0) {
                                if (sta_p->author_sm->Author_hs.state == PTK_sm_INITIALIZE)
                                        IEEE8021x_reassoc(priv_p, sta_p);
                                Author_key_timers_tick(priv_p, sta_p );
                        }
                } /* end of list_for_each */
        }

        /* If dot11RSNEnabled=0, we still check blocking because authmode
         * could be changed during traffic blocked.
         */
        if (priv_p->rsnBlockingTraffic &&
            check_60_sec(priv_p->MIC_block_timer)) {
                cmd_macPort_ctl(priv_p->netdev, bcstAddr, false);
                priv_p->MIC_block_timer = 0;
        }
#if 1//ONE_TIMER
        if ((priv_p->mib.dot11RSNEnabled)&&(priv_p->GroupKeyUpdate == true)){
                IEEE80211i_event(priv_p,NULL,GROUP_REKEY_TIME); // work
        } else {
                priv_p->GKRekeyTimer++;
                if (priv_p->GKRekeyTimer == priv_p->GroupKeyUpdateTimer){
                        priv_p->GroupKeyUpdate = true;
                }
        }
#endif
        set_sm_timer(priv_p); /* Periodically */
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
*
* \return
*/
/*----------------------------------------------------------------*/
void set_sm_timer(i2220mgmt_t *priv_p)
{
        mgmt_timer_t    *tmr = &priv_p->sm_timer;

        mgmt_stop_timer(tmr);
        mgmt_start_timer(tmr, SDL_TIMER_INTERVAL,
                     (mgmt_timeout_func) sm_timer_func,
                     (unsigned long) priv_p);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int linux_set_ssid(struct net_device *dev, param_ssid_t *data)
{
        i2220mgmt_t     *priv_p = dev->priv;
        int             ret;

        if ((ret = param_set_ssid(dev, data)) != 0) {
                return ret;
        }

        /* Re-calculate PSK key if we are in WPA-PSK mode */
        if ((priv_p->authMode == AuthMode_WPAPSK)&&(!priv_p->directPskValue)) {
                InitializeWPA_PSK(priv_p, 0);
        }

        return 0; /* success */
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int linux_set_m_ssid(struct net_device *dev, param_ssid_t *data)
{
#if SUPPORT_MSSID
        i2220mgmt_t     *priv_p = dev->priv;
        int             ssid_id = priv_p->param_ssid_index;
        int             ret;

        if ((ret = param_set_m_ssid(dev, data)) != 0) {
                return ret;
        }

        /* Re-calculate PSK key if we are in WPA-PSK mode */
        if (priv_p->authMode == AuthMode_WPAPSK &&
            priv_p->ap.ssid_list[ssid_id].enabled &&
            !priv_p->ap.ssid_list[ssid_id].directPskValue) {
                InitializeWPA_PSK(priv_p, ssid_id);
        }

        return 0; /* success */
#else
        return -1;
#endif
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int linux_set_m_enable(void *dev, int enable)
{
#if SUPPORT_MSSID
        i2220mgmt_t     *priv_p = kal_dev_priv((kal_net_device_t*)dev);
        int             ssid_id = priv_p->param_ssid_index;

        /* 0 is reserved for primary SSID only. And primary SSID is always enabled.
        */
        if (ssid_id > MAX_SSID -1 || ssid_id < 1) {
                return -1;
        }
        priv_p->ap.ssid_list[ssid_id].enabled = enable ? true : false;
        if (priv_p->authMode == AuthMode_WPAPSK &&
            priv_p->ap.ssid_list[ssid_id].enabled &&
            !priv_p->ap.ssid_list[ssid_id].directPskValue) {
                InitializeWPA_PSK(priv_p, ssid_id);
        }

        MlmeStart_request(priv_p);
        return 0;
#else
        return -1;
#endif
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int linux_set_auth_mode(struct net_device *dev, param_authMode_t newAuthMode)
{
        i2220mgmt_t     *priv_p = dev->priv;
        int             i;
        int             legacy_auth = AuthMode_OPEN;

        if (newAuthMode < AuthMode_OPEN || newAuthMode > AuthMode_WPAPSK)
                return -1;

        priv_p->WPA_PSK = FALSE;
        mib_import(dot11RSNEnabled) = FALSE;
        mib_import(dot11Pure1XEnabled) = FALSE;
        for (i = 0; i < AUTHENTICATION_SUITE_COUNT; i++) {
                mib_import(dot11RSNConfigAuthenticationSuitesTable[i]).
                        dot11RSNConfigAuthenticationSuiteEnabled = FALSE;
        }

        switch (newAuthMode) {
        case AuthMode_OPEN:
        case AuthMode_SHARED:
        case AuthMode_SWITCH:
                legacy_auth = (int) newAuthMode;
                break;
        case AuthMode_OPEN_1X:
                mib_import(dot11Pure1XEnabled) = TRUE;
                break;
        case AuthMode_WPA:
                mib_import(dot11RSNConfigAuthenticationSuitesTable[1]).
                        dot11RSNConfigAuthenticationSuiteEnabled = TRUE;
                mib_import(dot11RSNEnabled) = TRUE;
                mib_import(dot11Pure1XEnabled) = TRUE;
                randomGK(priv_p);
                break;
        case AuthMode_WPAPSK:
                mib_import(dot11RSNConfigAuthenticationSuitesTable[2]).
                        dot11RSNConfigAuthenticationSuiteEnabled = TRUE;
                mib_import(dot11RSNEnabled) = TRUE;
                priv_p->WPA_PSK = TRUE;
        #if SUPPORT_MSSID
                for (i = 0;i < MAX_SSID; i++) {
                        if (priv_p->ap.ssid_list[i].enabled &&
                            !priv_p->ap.ssid_list[i].directPskValue)
                                InitializeWPA_PSK(priv_p, i);
                }
        #else
                if (!priv_p->directPskValue)
                        InitializeWPA_PSK(priv_p, 0);
        #endif
                break;
        }

        rsnGenWPAIE(priv_p);
        param_set_authmode(dev, (param_authMode_t) legacy_auth);

        /* This statement should be after calling param_set_authmode(),
         * in which priv_p->authMode is set and MlmeStart_request()
         * is also invoked.
         */
        priv_p->authMode = newAuthMode;
        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int linux_set_cipher_mode(struct net_device *dev, param_cipherMode_t data)
{
        i2220mgmt_t     *priv_p = dev->priv;

        /* Update unicast and grooup cipher mode. In IPNRSN, unicast
         * and group cipher modes are the same. We permit the station
         * use group key as its pairwise cipher suite.
         */
        switch (data) {
        case CipherMode_NONE:   /* Disable WEP, TKIP, AES */
                set_wpa_uni_cipher_suite(priv_p, TRUE, FALSE, FALSE,
                        FALSE, FALSE, FALSE, FALSE);
                break;
        case CipherMode_WEP40:  /* Enable WEP. Disable TKIP, AES */
                set_wpa_uni_cipher_suite(priv_p, TRUE, TRUE,
                        FALSE, FALSE, FALSE, FALSE, FALSE);
                break;
        case CipherMode_WEP104: /* Enable WEP. Disable TKIP, AES */
                set_wpa_uni_cipher_suite(priv_p, TRUE, FALSE,
                        FALSE, FALSE, FALSE, TRUE, FALSE);
                break;
        case CipherMode_WEP128: /* Enable WEP. Disable TKIP, AES */
                set_wpa_uni_cipher_suite(priv_p, TRUE, FALSE,
                        FALSE, FALSE, FALSE, FALSE, TRUE);
                break;
        case CipherMode_TKIP:   /* Enable TKIP. Disable WEP, AES */
                set_wpa_uni_cipher_suite(priv_p, TRUE, FALSE,
                        TRUE, FALSE, FALSE, FALSE, FALSE);
                break;
        case CipherMode_CCMP:   /* Enable AES, Disable WEP, TKIP */
                set_wpa_uni_cipher_suite(priv_p, TRUE, FALSE,
                        FALSE, FALSE, TRUE, FALSE, FALSE);
                break;
        default:
                KDBG_LOG_INIT(KDBG_DEBUG, "Unknown cipherMode\n");
                return -1;
        } /* end of switch */

        /* Update cipher suite of group key */
        wlanoidSetEncryptionStatus(priv_p, data);

        /* Update WPA information element. In not IPNRSN, invoke
         * funtion cmd_set_RsnIE() to update RSN IE.
         */
        rsnGenWPAIE(priv_p);
        /* In MlmeStart_request(), all station records are cleaned up
         * except WDS. Four group keys ???
         */
        MlmeStart_request(priv_p);
        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int linux_set_passphrase(struct net_device *dev, unsigned char *data)
{
        i2220mgmt_t     *priv_p = dev->priv;
        int             i;

        if (strlen(data)==64) {
                for (i = 0; i < 64; i++) {
                        if (isdigit(data[i])){
                                data[i]-='0';
                        } else if (isupperhex(data[i])) {
                                data[i]-='A';
                                data[i]+=10;
                        } else {
                                KDBG_LOG_INIT(KDBG_DEBUG, "Error passphrase format !!!\n");
                                return -1;
                        }
                }
                for (i = 0; i < 64; i+=2)
                        data[i/2] = data[i+0] << 4 | (data[i+1]);
                for (i = 32; i < 64; i++)
                        data[i]=0x00;
        #if SUPPORT_MSSID
                for (i = 0; i < MAX_SSID; i++) {
                        strncpy(priv_p->ap.ssid_list[i].dot11RSNConfigPSKValue,
                                (UINT8*)data, 32);
                        priv_p->ap.ssid_list[i].directPskValue = TRUE;
                }
        #else
                        strncpy(mib_import(dot11RSNConfigTable).dot11RSNConfigPSKValue,
                                (UINT8*)data, 32);
                        priv_p->directPskValue = TRUE;
        #endif
                randomGK(priv_p);
                return 0;
        } else if (strlen(data) < 1 || strlen(data) > 63)
                return -1;
#if SUPPORT_MSSID
        for (i = 0; i < MAX_SSID; i++) {
                strncpy(priv_p->ap.ssid_list[i].dot11RSNConfigPSKPassPhrase,
                        (UINT8*)data, 64);
                priv_p->ap.ssid_list[i].dot11RSNConfigPSKPassPhrase[63] = 0;
                priv_p->ap.ssid_list[i].directPskValue = FALSE;
        }
#else
        strncpy(mib_import(dot11RSNConfigTable).dot11RSNConfigPSKPassPhrase,
                        (UINT8*)data, 64);
        mib_import(dot11RSNConfigTable).dot11RSNConfigPSKPassPhrase[63] = 0;
        priv_p->directPskValue = FALSE;
#endif

        if (priv_p->authMode == AuthMode_WPAPSK) {
        #if SUPPORT_MSSID
                for (i = 0; i < MAX_SSID; i++) {
                        if (priv_p->ap.ssid_list[i].enabled)
                                InitializeWPA_PSK(priv_p, i);
                }
        #else
                InitializeWPA_PSK(priv_p, 0);
        #endif
                MlmeStart_request(priv_p);
        }
        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int linux_set_GKRekeyTimer(struct net_device *dev, int data)
{
        i2220mgmt_t     *priv_p = dev->priv;

        if (data < 0)
                return -1;

        priv_p->GroupKeyUpdateTimer = data;
        priv_p->GKRekeyTimer = 0;
        return 0;
}

/*_________________________________________________________________________
**  set_wpa_uni_cipher_suite
**
**  descriptions: This routine will enable/disable the WPA unicast cipher
**                suite
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      NONE   - enable/disable for cipher NONE
**      WEP40  - enable/disable for cipher WEP40
**      TKIP   - enable/disable for cipher TKIP
**      WRAP   - enable/disable for cipher WRAP
**      CCMP   - enable/disable for cipher CCMP
**      WEP104 - enable/disable for cipher WEP104
**  return:
**  note:
**__________________________________________________________________________*/
void
set_wpa_uni_cipher_suite(IN PMP_ADAPTER adapter_p,
                                IN BOOLEAN none,
                                IN BOOLEAN wep40,
                                IN BOOLEAN tkip,
                                IN BOOLEAN wrap,
                                IN BOOLEAN ccmp,
                                IN BOOLEAN wep104,
                                IN BOOLEAN wep128)
{
        adapter_p->mib.dot11RSNConfigUnicastCiphersTable[0].
                                dot11RSNConfigUnicastCipherEnabled = none;
        adapter_p->mib.dot11RSNConfigUnicastCiphersTable[1].
                                dot11RSNConfigUnicastCipherEnabled = wep40;
        adapter_p->mib.dot11RSNConfigUnicastCiphersTable[2].
                                dot11RSNConfigUnicastCipherEnabled = tkip;
        adapter_p->mib.dot11RSNConfigUnicastCiphersTable[3].
                                dot11RSNConfigUnicastCipherEnabled = wrap;
        adapter_p->mib.dot11RSNConfigUnicastCiphersTable[4].
                                dot11RSNConfigUnicastCipherEnabled = ccmp;
        adapter_p->mib.dot11RSNConfigUnicastCiphersTable[5].
                                dot11RSNConfigUnicastCipherEnabled = wep104;
        adapter_p->mib.dot11RSNConfigUnicastCiphersTable[6].
                                dot11RSNConfigUnicastCipherEnabled = wep128;
}

#endif

#if SUPPORT_WPA2
void linux_set_11i_mode(struct net_device *dev, unsigned int data)
{
        i2220mgmt_t     *priv_p = kal_dev_priv(dev);

        mib_import(dot11RSNEnabled) = data;
}
#endif

/*----------------------------------------------------------------*/
/*!
* \brief WLAN_OS Interface between Linux and IPN2220AP driver
*
* \param
*
* \return
*/
/*----------------------------------------------------------------*/
static void indicate_to_1xsm(struct net_device* ap_dev, struct sk_buff* skb)
{
        char      dev1x[64];
        struct net_device *dev = NULL;

        if(!ap_dev){
                if(skb)
                        kal_free_skb(skb);
                printk(KERN_ALERT "Error!NULL AP device\n");
                return;
        }

        sprintf(dev1x,"%sap",ap_dev->name);
        dev = __dev_get_by_name(dev1x);
        if (skb && dev) {
                //linux_indicate_raw_data(dev, skb);
                skb->mac.raw = skb->data;
                skb->pkt_type = PACKET_OTHERHOST;
                memset(skb->cb, 0, sizeof(skb->cb));
                dev->last_rx = jiffies;

                /* Don't call kal_netdev_rx(), otherwise infinite loop */
                skb->protocol = eth_type_trans(skb, dev);
                skb->dev = dev; /* Important, crash if not correct */
                netif_rx(skb);
                return;
        } else {
                if(skb)
                        kal_free_skb(skb);
                printk(KERN_ALERT "Error!on indicate_to_1xsm\n");
                return;
        }

};

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int clone_and_forward_1x_port(i2220mgmt_t *priv_p, struct sk_buff *skb)
{
        struct sk_buff       *new_skb;

        if ((new_skb = (struct sk_buff*)kal_alloc_skb(skb->len)) != NULL) {
                /* Clone 802.1X frame to the user space daemon and
                 * 1x state machine.
                 */
                skb_put(new_skb, skb->len);
                memcpy(kal_skb_data(new_skb), kal_skb_data(skb), kal_skb_len(skb));
                indicate_to_1xsm((struct net_device*)priv_p->netdev, new_skb);
                return TRUE;
        }
	printk("clone_and_forward_1x_port: alloc skb fail\n");

        return FALSE;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int linux_handle_8021X(struct net_device *dev, struct sk_buff*skb)
{
        i2220mgmt_t             *priv_p = dev->priv;
        sta_info_t              *sta_p;
        wlan_mac_header_t       *wlan_p;
        ethhdr_t                eth;

        /* Transfer to 802.11 format from Ethernet format */
        memcpy(&eth, kal_skb_data(skb), sizeof(ethhdr_t));
        wlanhdr_from_ether(skb);        /* Fill LLC layer */
        kal_skb_push(skb, sMacHdrLng);  /* For 802.11 MAC header */

        /* Fill 802.11 header, wlan_p is 4-wise boundary here */
        wlan_p = (wlan_mac_header_t*) kal_skb_data(skb);
        wlan_p->frameCtrl = host2ieee16(SYN_data);
        sdl_setToDs(wlan_p, 1);
        memcpy(wlan_p->addr1, mib_import(dot11MACAddress), MAC_ADDR_LEN);
        memcpy(wlan_p->addr2, eth.h_source, MAC_ADDR_LEN);
        memcpy(wlan_p->addr3, eth.h_dest, MAC_ADDR_LEN);

        sta_p = task_sta_lookup(priv_p, wlan_p->addr2);
        if (!sta_p || skb->len < 34) {
                 goto L_freeskb;
        }

        /* Exclude ETH_P_PAE_KEY 1x frame */
#if SUPPORT_IPNRSN
        if (mib_import(dot11Pure1XEnabled) &&
            memcmp((UINT8*)wlan_p + 32, ETH_P_PAE_KEY, sizeof(ETH_P_PAE_KEY))) {
                /* Clone this packet to deamon */
                clone_and_forward_1x_port(priv_p, skb);
        }


        /* Check 802.1x EAPoL frame */
        if (mib_import(dot11RSNEnabled)) {
                if (!memcmp((UINT8*)wlan_p + 32, ETH_P_PAE_KEY,
                            sizeof(ETH_P_PAE_KEY))) {

                        ProcEapolKey(priv_p,
                                (EAPoL_KEY_MSG_t*)((UINT8*)wlan_p + 24),
                                sta_p->author_sm);
                }
                else if (!mib_import(dot11Pure1XEnabled) &&
                         !memcmp((UINT8*)wlan_p + 32,
                                ETH_P_PAE_START, sizeof(ETH_P_PAE_START))) {
                        /* EAPoL Start frame (Some vendors have bugs for rekey) */
                        sta_p->author_sm->Init = TRUE;
                }
        } /* end of process 1x frame */
#else
        if (mib_import(dot11Pure1XEnabled)) {
                /* Clone this packet to deamon */
                clone_and_forward_1x_port(priv_p, skb);
        }

#endif

L_freeskb:
        kal_free_skb(skb);
        return TRUE;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int kal_auth_request(void*net_dev, unsigned char *macAddr)
{
        struct sk_buff  *skb;
        sta_info_t* sta_p;
        struct net_device *dev = (struct net_device *)net_dev;
        i2220mgmt_t* priv_p = dev->priv;

#if WLAN_MACFILTER
	unsigned char* temp_macAddr = macAddr;
	int i,tbl_idx;
	//printk("Sam add debug 684 macfilter_type is %d\n",macfilter_tbl.macfilter_type);
	
	/* KenWang added 2005/06/21 -- begin -- */
	/* since we can't block 'eth0' in 6700, */
	/* we sould add accepted macs to iptables if br0 is blocked */
	struct task_struct *p;
	siginfo_t si;
	int pid = 0;
	char buffer[20];
	
	pid = fac_pid;
	si.si_signo = SIGUSR2;
	si.si_errno = 0;
	p = find_task_by_pid(pid);
	send_sig_info(SIGUSR2, &si, p);
	/* KenWang added 2005/06/21 -- end -- */
	
	if (macfilter_tbl.macfilter_type)
	{	
		int match = 0;
		for (tbl_idx=0; tbl_idx < macfilter_tbl.macfilter_count; tbl_idx++ )
		{
			macAddr = temp_macAddr;
			for (i=0; i < 6; i++, macAddr++)
			{
			//printk("%02x    %02x\n",*macAddr,macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[i]);
				if (*macAddr != macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[i])
				{
					match = 0;	
					break;
				}
				match = 1;
			}
			if (match)
				break;
		}
		macAddr = temp_macAddr; // KenWang added 2005/06/29
		if (macfilter_tbl.macfilter_type == SPECIFY_BLACK_LIST)
		{
			if (match)
			{
				/* KenWang added 2005/06/09 -- begin -- */
				/* add to security log if mac is in the filtering list */
				if(macfilter_tbl.log[tbl_idx])
					printk(KERN_DEBUG "WirelessFilter:Drop;SRC=%02X:%02X:%02X:%02X:%02X:%02X\n", 
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[0],
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[1],
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[2],
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[3],
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[4],
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[5]);
				//printk("BLACK_LIST match\n");
				/* KenWang added 2005/06/09 -- end -- */
				return 1;
			}
		}
		else if (macfilter_tbl.macfilter_type == SPECIFY_WHITE_LIST)
		{
			if (!match)
			{
				//printk("WHITE_LIST un-match\n");
				return 1;
			}
			else
			{
				/* KenWang added 2005/06/09 -- begin -- */
				/* add to security log if mac is in the filtering list */
				if(macfilter_tbl.log[tbl_idx])
					printk(KERN_DEBUG "WirelessFilter:Accept;SRC=%02X:%02X:%02X:%02X:%02X:%02X\n", 
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[0],
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[1],
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[2],
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[3],
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[4],
							macfilter_tbl.macaddr_array_ptr[tbl_idx].addr[5]);
				/* KenWang added 2005/06/09 -- end -- */
			}
		}
	}
	
	/* KenWang added 2005/06/27 -- begin -- */
	/* notify factoryreset to update iptable for wireless clients */
	send_sig_info(SIGUSR2, &si, p);
	/* KenWang added 2005/06/27 -- begin -- */
#endif

        sta_p = task_sta_lookup(priv_p, macAddr);
        #if SUPPORT_IPNRSN
        if(sta_p){
                memset(sta_p->wpaIE_p, 0 , sizeof(eREQ_WPA_t));
                Author_sm_alloc(priv_p,sta_p);
        }
        #endif
        if ((skb = kal_alloc_skb(24+6)) == NULL) {
		printk("kal_auth_request: alloc skb fail\n");
                return 0; /* accept */
	}

        kal_skb_put(skb, 24+6);

        /* Create auth frame with open system, seqNum=1 */
        *(unsigned short *)(skb->data) = host2chip_16(0x00b0);
        memcpy(skb->data + 4 , dev->dev_addr, 6);
        memcpy(skb->data + 10, macAddr, 6);
        memcpy(skb->data + 16, dev->dev_addr, 6);
        *(unsigned short *)(skb->data + 24) = host2chip_16(0); /* open */
        *(unsigned short *)(skb->data + 26) = host2chip_16(1); /* seqNum=1 */

        indicate_to_1xsm(dev, skb);

        return 0; /* accept */
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int kal_assoc_request(void *net_dev, unsigned char *macAddr,
                unsigned char *infoElem, int infoLen)
{
        struct net_device *dev = (struct net_device *)net_dev;
        i2220mgmt_t     *priv_p = dev->priv;
        sta_info_t      *sta_p;
        struct sk_buff  *skb;
#if SUPPORT_IPNRSN
        eREQ_WPA_t      *wpa_p;
        int             remainLen;
#endif
        if ((sta_p = task_sta_lookup(priv_p, macAddr)) != NULL) {
                sta_p->KeyMapOn = FALSE;
        }

#if SUPPORT_IPNRSN
        /* RSN enabled, check RSN information element */
        if (mib_import(dot11RSNEnabled))
        {
                remainLen = infoLen;
                wpa_p = (eREQ_WPA_t*)infoElem;
                do {
                        /* If remainLen<=0, we got NULL wpa_p */
                        wpa_p = (eREQ_WPA_t*) __search_infoElem(SYN_eWPA,
                                                        (UINT8*)wpa_p, remainLen);

                        /* No WPA element or match WPA OUI and version */
                        if (!wpa_p || (wpa_p->length >= 6 &&
                              !memcmp(&wpa_p->oui[0], IE_WPA_OUI, 4) &&
                              *(UINT8*)&wpa_p->version == 0x01)) {
                                break;
                        }
                        wpa_p = (eREQ_WPA_t*)((UINT8*)wpa_p + wpa_p->length + 2);
                        remainLen = infoLen - ((UINT32)wpa_p - (UINT32)infoElem);
                } while (1);

                if (wpa_p == NULL || sta_p == NULL ||
                    !rsnParseWPAIE(priv_p, sta_p, wpa_p)) {
                        KDBG_LOG_SDL(KDBG_DEBUG,"Not pass WPA checking.\n");
                        //printk(KERN_ALERT "Not pass WPA checking.\n");
                        return SYN_fail_other;
                }
                KDBG_LOG_SDL(KDBG_DEBUG, "Pass WPA checking\n");
                cmd_macPort_ctl(dev, macAddr, TRUE);
        }
#endif

        /* Check if this packet is passed to daemon */
        if (mib_import(dot11Pure1XEnabled)) {
                if ((skb = kal_alloc_skb(24+infoLen+10)) == NULL) {
			printk("kal_assoc_request: alloc skb fail\n");
                        return 0; /* accept */
		}

                kal_skb_put(skb, 24+infoLen+10);

                /* Create assoc_request frame  */
                *(unsigned short *)(skb->data) = host2chip_16(0x0020);
                memcpy(skb->data + 4 , dev->dev_addr, 6);
                memcpy(skb->data + 10, macAddr, 6);
                memcpy(skb->data + 16, dev->dev_addr, 6);
                memcpy(skb->data + 28, dev->dev_addr, 6);
                memcpy(skb->data + 34, infoElem, infoLen);
                indicate_to_1xsm(dev, skb);
                cmd_macPort_ctl(dev, macAddr, TRUE);
        }
#if SUPPORT_IPNRSN
        sta_p->author_sm->Init = TRUE;
        //IEEE8021x_reassoc(priv_p, sta_p);
#endif
        return SYN_sts_successful; /* Status in assoc_response frame */
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_disassoc_indicate(void*net_dev, unsigned char *macAddr)
{
#if SUPPORT_IPNRSN
        struct net_device* dev = (struct net_device*)net_dev;
        void* sta_p =NULL;
        sta_p = task_sta_lookup(dev->priv, macAddr);
        IEEE8021x_deauth(dev->priv, sta_p);
#endif
        return; /* Nothing to do in Linux */
}
/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_tkip_micFailure(void *net_dev, unsigned char *macAddr)
{
#if SUPPORT_IPNRSN
        struct net_device* dev = (struct net_device*)net_dev;
        i2220mgmt_t     *priv_p = dev->priv;
        sta_info_t      *sta_p = task_sta_lookup(priv_p, macAddr);

        if (!sta_p) return;
        sta_p->MIC_error_count++;

        if (sta_p->MIC_error_count == 1) { /* For the first Mic occur */
                sta_p->MIC_error_time = jiffies;
        }
        else if (sta_p->MIC_error_count >= 2) {
                if ( !check_60_sec(sta_p->MIC_error_time)) {
                        KDBG_LOG_TX(KDBG_WARN,"TKIP CounterMeasure occur!!\n");

                        /* Initialize the Countermeasure block traffic flag
                         * and timer, 60sec
                         */
                        IEEE80211i_event(priv_p, macAddr, DATA_MIC_FAILURE);

                        cmd_macPort_ctl(dev, bcstAddr, true);
                        priv_p->MIC_block_timer = jiffies;
                        sta_p->MIC_error_count = 0;
                }
                else {
                        /* The Second Mic error occure time is over 60 secs
                         * since first mic error occur
                         */
                        sta_p->MIC_error_time = jiffies;
                        sta_p->MIC_error_count = 1;
                }
        }
        return;
#else
        return;
#endif
}


