/*
** $Id: bci.c 1.20 2004/10/04 18:08:37 cmchang Exp $
*/

/*************************************************************************
* Copyright (c) 2002-2004 Inprocomm, Inc.
*
* All rights reserved. Copying, compilation, modification, distribution
* or any other use whatsoever of this material is strictly prohibited
* except in accordance with a Software License Agreement with
* Inprocomm, Inc.
**************************************************************************/
/*************************************************************************
GENERAL DESCRIPTION
********************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"

/************************************************************************
*                             M A C R O S
*************************************************************************
*/
#define BCI_WAIT_READY          1
#define BCI_LOOP_TIMES          5
#define BCI_CHK_OK_COUNT        10

/************************************************************************
*              Private Data
*************************************************************************
*/
DEBUGFILE("bci.c");

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/


/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*_________________________________________________________________________
**  BCIWrite8
**
**  descriptions: write 8-bit data in BCI interface
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      bciIndex - index of BB CR
**      bciData - value write to BB CR
**  return: TRUE - write success
**          FALSE - write fail
**  note:
**__________________________________________________________________________*/
BOOLEAN
BCIWrite8(IN PMP_ADAPTER adapter_p,
          IN UINT_8 bciIndex,
          IN UINT_8 bciData,
          IN BOOLEAN requireCHK,
          IN UINT_8 chkMask)
{
    UINT_32 pbarData;
    BOOLEAN result;
    UINT_32 tmp32;
#if BCI_WORKAROUND
    UINT i;
    UINT_8 readData;
#endif

    DEBUGFUNC("BCIWrite8");

	ASSERT(bciIndex <= 127);

    pbarData = PBAR_PHY_BCI_ACCESS_START | PBAR_PHY_BCI_RW_CTRL_WRITE |
        (((UINT_32) bciIndex) << 16) | (((UINT_32) bciData) << 8);

#if BCI_WORKAROUND
    for (i = 0; i < BCI_CHK_OK_COUNT; i++) {
	    NIC_MCR_READ_AND_WAIT(adapter_p, MCR_PBAR, &tmp32,
	        (tmp32 & PBAR_PHY_BCI_ACCESS_START) == 0, BCI_WAIT_READY,
	        BCI_LOOP_TIMES, result);

	    if (!result) {
	        WARNLOG(("Cannot write CR%d -- BCI busy\n", bciIndex));
	        continue;
	    }

        NIC_MCR_WRITE(adapter_p, MCR_PBAR, pbarData);

        NIC_MCR_READ_AND_WAIT(adapter_p, MCR_PBAR, &tmp32,
            (tmp32 & PBAR_PHY_BCI_ACCESS_START) == 0, BCI_WAIT_READY,
            BCI_LOOP_TIMES, result);

        if (!requireCHK) {
#if DBG
            if (!result) {
                WARNLOG(("Write CR%d -- not complete soon\n", bciIndex));
            }
#endif
            break;
        }

        result = BCIRead8(adapter_p, bciIndex, &readData, FALSE, 0x00);

        if (!result) {
            continue;
        }

        if ((readData & chkMask) == (bciData & chkMask)) {
            break;
        }
    }
#else
    do {
	    NIC_MCR_READ_AND_WAIT(adapter_p, MCR_PBAR, &tmp32,
	        (tmp32 & PBAR_PHY_BCI_ACCESS_START) == 0, BCI_WAIT_READY,
	        BCI_LOOP_TIMES, result);

	    if (!result) {
	        WARNLOG(("Cannot write CR%d -- BCI busy\n", bciIndex));
	        break;
	    }

        NIC_MCR_WRITE(adapter_p, MCR_PBAR, pbarData);

        NIC_MCR_READ_AND_WAIT(adapter_p, MCR_PBAR, &tmp32,
            (tmp32 & PBAR_PHY_BCI_ACCESS_START) == 0, BCI_WAIT_READY,
            BCI_LOOP_TIMES, result);

        if (!result) {
            WARNLOG(("Write CR%d -- not complete soon\n", bciIndex));
        }
    } while (FALSE);
#endif

#if DBG
    if (!result) {
        ERRORLOG(("BCI write CR%d = %x failed\n", bciIndex, bciData));
    }
#endif

    return result;
}   /* BCIWrite8 */

/*_________________________________________________________________________
**  BCIRead8
**
**  descriptions: read 8-bit data in BCI interface
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      bciIndex - index of BB CR
**      bciData_p - pointer to fill read value from BB CR
**  return: TRUE -> read success
**          FALSE -> read fail
**  note:
**__________________________________________________________________________*/
BOOLEAN
BCIRead8(IN PMP_ADAPTER adapter_p,
         IN UINT_8 bciIndex,
         OUT PUINT_8 bciData_p,
         IN BOOLEAN requireCHK,
         IN UINT_8 chkMask)
{
    UINT_32 pbarData;
    UINT_8 readData = 0;
    BOOLEAN result;
    UINT_32 tmp32;
#if BCI_WORKAROUND
    UINT i;
    BOOLEAN successfulRead = FALSE;
    UINT_8 tmp8;
#endif

    DEBUGFUNC("BCIRead8");

	ASSERT(bciIndex <= 127);

    pbarData = PBAR_PHY_BCI_ACCESS_START | PBAR_PHY_BCI_RW_CTRL_READ |
        (((UINT_32) bciIndex) << 16);

#if BCI_WORKAROUND
    for (i = 0; i < BCI_CHK_OK_COUNT; i++) {
	    NIC_MCR_READ_AND_WAIT(adapter_p, MCR_PBAR, &tmp32,
	        (tmp32 & PBAR_PHY_BCI_ACCESS_START) == 0, BCI_WAIT_READY,
	        BCI_LOOP_TIMES, result);

	    if (!result) {
	        continue;
	    }

        NIC_MCR_WRITE(adapter_p, MCR_PBAR, pbarData);

        NIC_MCR_READ_AND_WAIT(adapter_p, MCR_PBAR, &tmp32,
            (tmp32 & PBAR_PHY_BCI_ACCESS_START) == 0, BCI_WAIT_READY,
            BCI_LOOP_TIMES, result);

        if (!requireCHK) {
            if (result) {
                readData = (UINT_8) (tmp32 & PBAR_PHY_BCI_RD_DATA);
            }
#if DBG
            else {
                WARNLOG(("Read CR%d -- not complete soon\n", bciIndex));
            }
#endif
            break;
        }

        if (!result) {
            continue;
        }

        tmp8 = (UINT_8) (tmp32 & PBAR_PHY_BCI_RD_DATA);

        if (successfulRead) {
            if ((readData & chkMask) == (tmp8 & chkMask)) {
                break;
            }
        }

        successfulRead = TRUE;
        readData = tmp8;
    }
#else
    do {
	    NIC_MCR_READ_AND_WAIT(adapter_p, MCR_PBAR, &tmp32,
	        (tmp32 & PBAR_PHY_BCI_ACCESS_START) == 0, BCI_WAIT_READY,
	        BCI_LOOP_TIMES, result);

	    if (!result) {
	        WARNLOG(("Cannot read CR%d -- BCI busy\n", bciIndex));
	        break;
	    }

        NIC_MCR_WRITE(adapter_p, MCR_PBAR, pbarData);

        NIC_MCR_READ_AND_WAIT(adapter_p, MCR_PBAR, &tmp32,
            (tmp32 & PBAR_PHY_BCI_ACCESS_START) == 0, BCI_WAIT_READY,
            BCI_LOOP_TIMES, result);

        if (!result) {
            WARNLOG(("Read CR%d -- not complete soon\n", bciIndex));
        }

        readData = (UINT_8) (tmp32 & PBAR_PHY_BCI_RD_DATA);
    } while (FALSE);
#endif

    if (result) {
        *bciData_p = readData;
    } else {
        ERRORLOG(("BCI read CR%d failed\n", bciIndex));
        *bciData_p = 0xFF;
    }

    return result;
}   /* BCIRead8 */

/*_________________________________________________________________________
**  bciRead16
**
**  descriptions: This routine is called to read 16-bit data from the
**                Baseband through the BCI interface.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      bciIndexHigh - index of BB CR to be read as the higher byte
**      bciIndexLow - index of BB CR to be read as the lower byte
**      bciData_p - pointer to fill read value from BB CRs
**  return:
**      TRUE - read success
**      FALSE - read fail
**  note:
**__________________________________________________________________________*/
BOOLEAN
bciRead16(IN PMP_ADAPTER adapter_p,
          IN UINT_8 bciIndexHigh,
          IN UINT_8 bciIndexLow,
          OUT PUINT_16 bciData_p,
          IN BOOLEAN checkRequired,
          IN UINT_16 checkMask)
{
    BOOLEAN result;
    UINT_8 readData0, readData1;

    DEBUGFUNC("bciRead16");

    do {
        result = BCIRead8(adapter_p, bciIndexHigh, &readData1, checkRequired,
            (UINT_8) (checkMask >> 8));

        if (!result) {
            ERRORLOG(("Failed to read CR%d for higher byte\n", bciIndexHigh));
            break;
        }

        result = BCIRead8(adapter_p, bciIndexLow, &readData0, checkRequired,
            (UINT_8) (checkMask & 0x00FF));

        if (!result) {
            ERRORLOG(("Failed to read CR%d for lower byte\n", bciIndexLow));
            break;
        }

        DBGLOG(HW, INFO, ("Read succeeded - (CR%d, CR%d)=(%d, %d)\n",
            bciIndexHigh, bciIndexLow, readData1, readData0));
    } while (FALSE);

    if (result) {
        *bciData_p = ((UINT_16) readData1 << 8) | (UINT_16) readData0;
    } else {
        *bciData_p = 0xFFFF;
    }

    return result;
}   /* bciRead16 */

