/*
** $Id: eeprom.c 1.20 2004/10/04 18:10:10 cmchang Exp $
*/

/*************************************************************************
* Copyright (c) 2002-2004 Inprocomm, Inc.
*
* All rights reserved. Copying, compilation, modification, distribution
* or any other use whatsoever of this material is strictly prohibited
* except in accordance with a Software License Agreement with
* Inprocomm, Inc.
**************************************************************************/
/*************************************************************************
GENERAL DESCRIPTION
********************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"


/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/
typedef struct _EEPROMLIST_STRUC {
	UINT_16  offset;
	UINT_16  initValue;
} EEPROMLIST_STRUC, *PEEPROMLIST_STRUC;


/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/
DEBUGFILE("eeprom.c");


/************************************************************************
*                             M A C R O S
*************************************************************************
*/
#define TTL_ACCESS_RDY_MSEC     20
#define MAX_LOOP_ACCESS_RDY     200


/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/
UINT_8
VerifyEE_CSERR(IN PMP_ADAPTER adapter_p);


/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*_________________________________________________________________________
**  WriteEEProm16
**
**  descriptions: write 16-bit data to EEPRom
**  parameters:
**  return:
**  note:
**__________________________________________________________________________*/
UINT_8
WriteEEProm16(IN PMP_ADAPTER adapter_p,
              IN UINT_8 EEPromOffset,
              IN UINT_16 EEPromData)
{
    BOOLEAN bReady;
    UINT_32 wdValue = ( ((UINT_32) EEPromOffset) << 16) + EEPromData;
    UINT_32 regValue;

    DEBUGFUNC("WriteEEProm16");

    DBGLOG(HW, INFO, ("ReadEEProm16 >> wdValue:0x%lx\n", wdValue));

    NIC_MCR_WRITE(adapter_p, MCR_ECSR, ECSR_EE_RDY);

    NIC_MCR_WRITE(adapter_p, MCR_EADR, wdValue);
    NIC_MCR_WRITE(adapter_p, MCR_ERWCR, ERWCR_EE_WRITE);

    NIC_MCR_READ_AND_WAIT(adapter_p, MCR_ECSR, &regValue,
        (regValue & ECSR_EE_RDY) == ECSR_EE_RDY, TTL_ACCESS_RDY_MSEC,
        MAX_LOOP_ACCESS_RDY, bReady);

    if (bReady) {
        NIC_MCR_WRITE(adapter_p, MCR_ECSR, ECSR_EE_RDY);
    } else {
        ERRORLOG(("EEPROM write fail due to time out\n"));
    }

    return ((UINT_8) bReady);
}   /* WriteEEProm16 */

/*_________________________________________________________________________
**  ReadEEProm16
**
**  descriptions: read 16-bit data to EEPROM
**  parameters:
**  return:
**  note:
**__________________________________________________________________________*/
UINT_16
ReadEEProm16(IN PMP_ADAPTER adapter_p,
             IN UINT_8 EEPromOffset)
{
    BOOLEAN bReady;
    UINT_32 wdValue = ((UINT_32) EEPromOffset) << 16;
    UINT_32 regValue;

    DEBUGFUNC("ReadEEProm16");

    DBGLOG(HW, INFO, ("ReadEEProm16 >> wdValue:0x%x\n", wdValue));

    NIC_MCR_WRITE(adapter_p, MCR_ECSR, ECSR_EE_RDY);

    NIC_MCR_WRITE(adapter_p, MCR_EADR, wdValue);
    NIC_MCR_WRITE(adapter_p, MCR_ERWCR, ERWCR_EE_READ);

    NIC_MCR_READ_AND_WAIT(adapter_p, MCR_ECSR, &regValue,
        (regValue & ECSR_EE_RDY) == ECSR_EE_RDY, TTL_ACCESS_RDY_MSEC,
        MAX_LOOP_ACCESS_RDY, bReady);

    if (bReady) {
        NIC_MCR_WRITE(adapter_p, MCR_ECSR, ECSR_EE_RDY);
        NIC_MCR_READ(adapter_p, MCR_EADR, &regValue);
        DBGLOG(HW, INFO, ("MCR_EADR: 0x%x\n", regValue));
    } else {
        ERRORLOG(("EEPROM read fail due to time out\n"));
    }

    return ((UINT_16) regValue);
}   /* ReadEEProm16 */

/*_________________________________________________________________________
**  EraseEEPRom
**
**  descriptions: Flush EEPRom (write EEPRom to whole 0xffff)
**  parameters:
**  return:
**  note:
**__________________________________________________________________________*/
VOID
EraseEEPRom(IN PMP_ADAPTER adapter_p)
{
    UINT_8 i;

    for(i = 0; i < (512 / 2); i++){
        WriteEEProm16(adapter_p, i, 0xffff);
    }
}   /* EraseEEPRom */

/*_________________________________________________________________________
**  VerifyEE_CSERR
**
**  descriptions: check if the EE_CSERR in offset 0x14 register is set
**  parameters:
**  return:
**  note:
**__________________________________________________________________________*/
UINT_8
VerifyEE_CSERR(IN PMP_ADAPTER adapter_p)
{
    UINT_32 wReadValue;
    NIC_MCR_READ(adapter_p, MCR_ECSR, &wReadValue);
    wReadValue = (wReadValue & 0x08 ) >> 3;

    return (UINT_8) wReadValue;
}   /* VerifyEE_CSERR */

/*_________________________________________________________________________
**  VerifyCheckSum
**
**  descriptions: calculate checksum of EEPRom from word offset 0x01 ~ 0x0E
**                with 0xff, and check EE_CSERR in offset 0x14 register also
**  parameters:
**  return:
**  note:
**__________________________________________________________________________*/
UINT_8
VerifyCheckSum(IN PMP_ADAPTER adapter_p)
{
    UINT_8   wd;
    UINT_16 wChkSum = 0;
    UINT_16 wReadValue;

    for(wd = 1; wd <= 0x0F; wd++){
        wReadValue = ReadEEProm16(adapter_p, wd);
        wChkSum += ((wReadValue & 0xff) + (wReadValue >> 8));
    }

    if( (wChkSum & 0x0ff) == 0x0ff ){
        return 0xff;
    } else {
        if( VerifyEE_CSERR(adapter_p) == 0){
            WARNLOG(("Check sum error but EE_CSERR does not set...\n\r"));
        }

        return (UINT_8) wChkSum;
    }
}   /* VerifyCheckSum */

